# DragonLog (c) 2023-2025 by Andreas Schawo is licensed under CC BY-SA 4.0.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/

import logging

from PyQt6 import QtWidgets, QtCore
from hamcc import hamcc

from . import CassiopeiaConsole_ui
from .Logger import Logger
from .RegEx import check_call
from .contest import CONTESTS, CONTEST_IDS, CONTEST_NAMES, ExchangeData, ContestLog
from .cty import Country
from .distance import distance
from .local_callbook import LocalCallbook


# noinspection PyPep8Naming
class CassiopeiaConsole(QtWidgets.QDialog, CassiopeiaConsole_ui.Ui_CassiopeiaConsoleWidget):
    qsosChached = QtCore.pyqtSignal()

    def __init__(self, parent, dragonlog, settings: QtCore.QSettings, logger: Logger, local_cb: LocalCallbook):
        super().__init__(parent)
        self.dragonlog = dragonlog
        self.setupUi(self)

        self.log = logging.getLogger('CassiopeiaConsole')
        self.log.addHandler(logger)
        self.log.setLevel(logger.loglevel)
        self.logger = logger
        self.log.debug('Initialising...')
        self.log.debug(f'HamCC version: {hamcc.__version_str__}...')

        self.__settings__ = settings
        self.__local_cb__ = local_cb

        self.__cc__ = hamcc.CassiopeiaConsole(self.__settings__.value('station/callsign', ''),
                                              self.__settings__.value('station/qth_loc', ''),
                                              self.__settings__.value('station/name', ''))
        self.__current_call__ = ''
        self.__current_rxexch__ = ''
        self.__cdata__ = None

        self.__initWidgets__()

        self.refreshDisplay()

    def __initWidgets__(self):
        bands: list = self.__settings__.value('ui/show_bands', [])
        if '11m' in bands:
            bands.remove('11m')
        self.bandComboBox.insertItems(1, bands)
        self.modeComboBox.insertItems(1, self.__settings__.value('ui/show_modes', []))
        self.eventComboBox.insertItems(3, CONTEST_IDS.keys())
        self.ctyCtyLabel.setText(f'? ?, ?')
        self.ctyAreaLabel.setText(f'DXCC=?, CQ=?, ITU=?')

    def _check_worked_before_(self, call):
        if check_call(call):
            worked_dict = self.dragonlog.workedBefore(call)
            worked_list = []
            if worked_dict:
                for call, data in zip(worked_dict.keys(), worked_dict.values()):
                    if self.__cc__.__event__ and data['event'] != self.__cc__.__event__:
                        continue
                    worked_list.append(f'{call} {self.tr("at")} {data["date_time"]}')
                    if len(worked_list) == 3:
                        break
            if worked_list:
                self.resultWidget.setStyleSheet('#resultWidget {background-color: rgba(0, 0, 255, 63)}')
                self.resultLabel.setText('\n'.join(worked_list))

    def _callbook_lookup_(self, call: str, contest: str = ''):
        self._lookup_cty_data_(call)

        if not self.__local_cb__:
            return

        if contest and self.__local_cb__.history_entries[0]:
            # Initialise
            exc_data = ExchangeData()
            if self.__cdata__:
                exc_data.itu_zone = self.__cdata__.itu

            ch_data = self.__local_cb__.lookup_history(contest, call, True)
            if ch_data:
                # Initialise from callbook
                if ch_data.callbook_data:
                    if ch_data.callbook_data.name:
                        self.evaluate(f'\'{ch_data.callbook_data.name} ')
                    if ch_data.callbook_data.locator:
                        exc_data.locator = ch_data.callbook_data.locator
                    if ch_data.callbook_data.darc_dok:
                        exc_data.darc_dok = ch_data.callbook_data.darc_dok

                # Overwrite from history if available
                if ch_data.history_data:
                    if ch_data.history_data.locator:
                        exc_data.locator=ch_data.history_data.locator
                    exc_data.power=ch_data.history_data.power_class
                    if ch_data.history_data.darc_dok:
                        exc_data.darc_dok=ch_data.history_data.darc_dok
                    if ch_data.history_data.itu_zone:
                        exc_data.itu_zone=ch_data.history_data.itu_zone

            contest = CONTESTS.get(contest, None)
            self.evaluate(f'%{contest.prepare_exchange(exc_data)} ')
        else:
            cb_data = self.__local_cb__.lookup(call, True)
            if cb_data:
                qso = self.__cc__.current_qso
                if cb_data.callbook_data.qth and cb_data.callbook_data.locator and not 'QTH' in qso:
                    self.evaluate(f'@{cb_data.callbook_data.qth} ({cb_data.callbook_data.locator}) ')
                elif cb_data.callbook_data.locator and not qso.get('GRIDSQUARE', ''):
                    self.evaluate(f'@{cb_data.callbook_data.locator} ')

                if cb_data.callbook_data.name and not qso.get('NAME', ''):
                    self.evaluate(f'\'{cb_data.callbook_data.name} ')

    def _lookup_cty_data_(self, call):
        if call:
            self.__cdata__: Country = self.dragonlog.cty_data(call)
            if self.__cdata__:
                self.ctyCtyLabel.setText(f'{self.__cdata__.code} {self.__cdata__.name}, {self.__cdata__.continent}')
                self.ctyAreaLabel.setText(
                    f'DXCC={self.__cdata__.dxcc}, CQ={self.__cdata__.cq}, ITU={self.__cdata__.itu}')
        else:
            self.__cdata__ = None
            self.ctyCtyLabel.setText(f'? ?, ?')
            self.ctyAreaLabel.setText(f'DXCC=?, CQ=?, ITU=?')

    def refreshDisplay(self):
        qso = self.__cc__.current_qso

        self.myCallLineEdit.setText(qso.get('STATION_CALLSIGN', ''))
        my_loc = f'{qso["MY_CITY"]} ({qso.get("MY_GRIDSQUARE", "")})' if 'MY_CITY' in qso else qso.get('MY_GRIDSQUARE',
                                                                                                       '')
        self.myLocLineEdit.setText(my_loc)
        self.myNameLineEdit.setText(qso.get('MY_NAME', ''))

        if 'CONTEST_ID' in qso:
            self.eventComboBox.setCurrentText(CONTEST_NAMES.get(qso['CONTEST_ID'], qso['CONTEST_ID']))
            self.exchTXLineEdit.setText(qso.get('STX_STRING', qso.get('STX', '')))
            self.exchRXLineEdit.setText(qso.get('SRX_STRING', qso.get('SRX', '')))

            contest: ContestLog | None = CONTESTS.get(qso['CONTEST_ID'], None)
            if contest:
                self.exchRXLineEdit.setPlaceholderText(contest.contest_exch_fmt)

                if not qso.get('SRX_STRING', ''):  # Reset exchange cache
                    self.__current_rxexch__ = ''
                elif qso.get('SRX_STRING', '') != self.__current_rxexch__:  # Only run if exchange changed
                    self.__current_rxexch__ = qso.get('SRX_STRING', '')
                    exch: ExchangeData = contest.extract_exchange(qso.get('SRX_STRING', ''))
                    if exch and exch.locator:
                        self.evaluate(f'@{exch.locator} ')
        elif 'MY_SIG' in qso:
            self.exchRXLineEdit.setPlaceholderText(self.tr('Exchange'))
            self.eventComboBox.setCurrentText(qso['MY_SIG'])
            self.exchTXLineEdit.setText(qso.get('MY_SIG_INFO', ''))
            self.exchRXLineEdit.setText(qso.get('SIG_INFO', ''))
        else:
            self.exchRXLineEdit.setPlaceholderText(self.tr('Exchange'))
            self.exchTXLineEdit.setText('')
            self.exchRXLineEdit.setText('')
            self.eventComboBox.setCurrentText(self.tr('Event ID'))

        self.onlineCheckBox.setChecked(self.__cc__.is_online())
        self.dateEdit.setDate(QtCore.QDate.fromString(qso['QSO_DATE'][:8], 'yyyyMMdd'))
        self.timeEdit.setTime(QtCore.QTime.fromString(qso['TIME_ON'][:4], 'HHmm'))
        self.dateEdit.setDisabled(self.onlineCheckBox.isChecked())
        self.timeEdit.setDisabled(self.onlineCheckBox.isChecked())
        self.dateEdit.blockSignals(self.onlineCheckBox.isChecked())
        self.timeEdit.blockSignals(self.onlineCheckBox.isChecked())

        self.bandComboBox.setCurrentText(qso['BAND'].lower() if qso['BAND'] else self.tr('Band'))
        self.modeComboBox.setCurrentText(qso['MODE'].upper() if qso['MODE'] else self.tr('Mode'))
        self.callLineEdit.setText(qso['CALL'])
        loc = f'{qso["QTH"]} ({qso["GRIDSQUARE"]})' if 'QTH' in qso else qso.get('GRIDSQUARE', '')
        self.locLineEdit.setText(loc)

        self.rstRXLineEdit.setText(qso.get('RST_RCVD', ''))
        self.rstTXLineEdit.setText(qso.get('RST_SENT', ''))
        self.nameLineEdit.setText(qso.get('NAME', ''))

        self.freqSpinBox.setValue(int(float(qso.get("FREQ", "0")) * 1000))
        self.powerSpinBox.setValue(int(qso.get('TX_PWR', '0')))
        self.qslCheckBox.setChecked(qso.get('QSL_RCVD', 'N') == 'Y')
        self.commentLineEdit.setText(qso.get('COMMENT', ''))

        call = self.__cc__.current_qso.get('CALL', '')
        if call != self.__current_call__:
            self.__current_call__ = call
            self._check_worked_before_(call)
            self._callbook_lookup_(call, self.__cc__.current_qso.get('CONTEST_ID', ''))

        if qso.get('GRIDSQUARE', '') and qso.get('MY_GRIDSQUARE', ''):
            # noinspection PyBroadException
            try:
                self.distLabel.setText(f'{distance(qso["GRIDSQUARE"], qso["MY_GRIDSQUARE"])} km')
            except Exception:
                self.distLabel.setText('? km')
        else:
            self.distLabel.setText('? km')

    def evaluate(self, text: str):
        if text.endswith('~'):
            self.clearInput()
        elif text.endswith('!'):
            self.inputLineEdit.setText(self.inputLineEdit.text()[:-1])
            self.inputLineEdit.clear()
            self.finaliseQSO()
            self.qsosChached.emit()
        elif text.endswith('?'):
            qso = str(self.__cc__.current_qso)
            for c in '{}\'':
                qso = qso.replace(c, '')
            self.resultLabel.setText(qso)
            self.resultWidget.setStyleSheet('#resultWidget {background-color: rgba(0, 255, 0, 63)}')
            self.inputLineEdit.clear()
        elif text.startswith('"'):
            if text.endswith('"'):
                self._evaluate_(text[1:-1])
        elif text.endswith(' '):
            self._evaluate_(text.strip())

        self.refreshDisplay()

    def setQSO(self, call: str, band: str = '', freq: float = .0):
        self.evaluate(f'{call} ')

        if band:
            self.evaluate(f'{band} ')
        if freq:
            freq_s = f'{freq:0.3f}'.rstrip('0').rstrip('.')
            self.evaluate(f'{freq_s}f ')

    def setFrequency(self, freq: float):
        freq_s = f'{freq:0.3f}'.rstrip('0').rstrip('.')
        self.evaluate(f'{freq_s}f ')

    def setBand(self, band: str):
        self.evaluate(f'{band} ')

    def setMode(self, mode: str):
        self.evaluate(f'{mode} ')

    def setPower(self, pwr: int):
        self.evaluate(f'{int(pwr)}p ')

    def _evaluate_(self, text):
        if text:
            res = self.__cc__.evaluate(text)
            self.resultLabel.setText(self.tr(res))
            if res.startswith('Error:'):
                self.resultWidget.setStyleSheet('#resultWidget {background-color: rgba(255, 0, 0, 63)}')
            elif res.startswith('Warning:'):
                self.resultWidget.setStyleSheet('#resultWidget {background-color: rgba(255, 127, 0, 63)}')
            else:
                self.resultWidget.setStyleSheet('#resultWidget {background-color: rgba(0, 255, 0, 63)}')
            self.inputLineEdit.clear()

    def finaliseQSO(self):
        text = self.inputLineEdit.text()
        if text.startswith('"'):
            self._evaluate_(text[1:])
        else:
            self._evaluate_(text)

        if self.__cc__.current_qso.get('CALL', ''):
            res = self.__cc__.finalize_qso()
            self.resultLabel.setText(self.tr(res))
            if res.startswith('Error:'):
                self.resultWidget.setStyleSheet('#resultWidget {background-color: rgba(255, 0, 0, 63)}')
            elif res.startswith('Warning:'):
                self.resultWidget.setStyleSheet('#resultWidget {background-color: rgba(255, 127, 0, 63)}')
            else:
                self.resultWidget.setStyleSheet('#resultWidget {background-color: rgba(0, 255, 0, 63)}')

            self.clearInput()
            self.qsosChached.emit()
        else:
            self.resultWidget.setStyleSheet('#resultWidget {background-color: rgba(255, 127, 0, 63)}')
            self.resultLabel.setText(self.tr('Error: Callsign missing for QSO'))

    def hasQSOs(self) -> bool:
        return self.__cc__.has_qsos()

    def retrieveQSO(self) -> dict[str, str]:
        qso = self.__cc__.pop_qso()
        return qso

    def clearInput(self):
        self.__cc__.clear()
        self.__current_call__ = ''
        self.inputLineEdit.clear()
        self.refreshDisplay()
        self.inputLineEdit.setFocus()

    def resetUserData(self):
        self._evaluate_('-c' + self.__settings__.value('station/callsign', ''))
        self._evaluate_('-l' + self.__settings__.value('station/qth_loc', ''))
        self._evaluate_('-n' + self.__settings__.value('station/name', ''))
        self.refreshDisplay()

    def tr(self, text: str, **kwargs) -> str:
        if text.startswith('Last QSO cached:'):
            text, call = text.split(':')
            return f'{super().tr(text)}:{call}'
        else:
            return super().tr(text)

    def __(self):
        # Stub for translations
        self.tr('Warning: Wrong call format')
        self.tr('Error: Wrong call format')
        self.tr('Error: Wrong QTH/maidenhead format')
        self.tr('Error: No active event')
        self.tr('Error: Unknown prefix')
        self.tr('Error: Wrong RST format')
        self.tr('Warning: Callsign missing for last QSO')
        self.tr('Last QSO cached')

    def callChanged(self, call: str):
        if len(call) > 2:
            self.evaluate(call + ' ')

    def locatorChanged(self):
        self.evaluate(f'@{self.locLineEdit.text()} ')

    def nameChanged(self):
        self._evaluate_('\'' + self.nameLineEdit.text())

    def commentChanged(self):
        self._evaluate_('#' + self.commentLineEdit.text())

    def bandChanged(self, band: str):
        if band and band != self.tr('Band'):
            self._evaluate_(band)
        self.refreshDisplay()

    def modeChanged(self, mode: str):
        if mode and mode != self.tr('Mode'):
            self._evaluate_(mode)
        self.refreshDisplay()

    def dateChanged(self, date: QtCore.QDate):
        self.evaluate(date.toString('yyyyMMdd') + 'd ')

    def timeChanged(self, time: QtCore.QTime):
        self.evaluate(time.toString('HHmm') + 't ')

    def freqChanged(self, freq: int):
        self._evaluate_(f'{freq}f')

    def powerChanged(self, power: int):
        self._evaluate_(f'{power}p')

    def qslChanged(self, state: bool):
        self._evaluate_('*')

    def rstRxChanged(self):
        if self.rstRXLineEdit.text():
            self._evaluate_('.' + self.rstRXLineEdit.text())

    def rstTxChanged(self):
        if self.rstTXLineEdit.text():
            self._evaluate_(',' + self.rstTXLineEdit.text())

    def myCallChanged(self):
        self._evaluate_('-c' + self.myCallLineEdit.text())

    def myLocChanged(self):
        self.evaluate(f'-l{self.myLocLineEdit.text()} ')

    def myNameChanged(self):
        self._evaluate_('-n' + self.myNameLineEdit.text())

    def eventChanged(self, event: str):
        if event and event != self.tr('Event ID'):
            event = CONTEST_IDS.get(event, self.eventComboBox.currentText().strip())
            self._evaluate_('$' + event)
        else:
            self._evaluate_('$')
        self.refreshDisplay()

    def exchRxChanged(self):
        event = self.eventComboBox.currentText()
        if event and event != self.tr('Event ID'):
            self._evaluate_('%' + self.exchRXLineEdit.text())

    def exchTxChanged(self):
        event = self.eventComboBox.currentText()
        if event and event != self.tr('Event ID'):
            self._evaluate_('-N' + self.exchTXLineEdit.text())

    def onlineChanged(self, _):
        self.evaluate('-o ')



# Form implementation generated from reading ui file 'ListEdit.ui'
#
# Created by: PyQt6 UI code generator 6.9.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_ListEditForm(object):
    def setupUi(self, ListEditForm):
        ListEditForm.setObjectName("ListEditForm")
        ListEditForm.resize(155, 180)
        self.verticalLayout = QtWidgets.QVBoxLayout(ListEditForm)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(parent=ListEditForm)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.listWidget = QtWidgets.QListWidget(parent=ListEditForm)
        self.listWidget.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.AnyKeyPressed|QtWidgets.QAbstractItemView.EditTrigger.DoubleClicked|QtWidgets.QAbstractItemView.EditTrigger.EditKeyPressed)
        self.listWidget.setAlternatingRowColors(True)
        self.listWidget.setObjectName("listWidget")
        self.verticalLayout.addWidget(self.listWidget)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.addPushButton = QtWidgets.QPushButton(parent=ListEditForm)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("icons:edit_add.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.addPushButton.setIcon(icon)
        self.addPushButton.setObjectName("addPushButton")
        self.horizontalLayout.addWidget(self.addPushButton)
        self.delPushButton = QtWidgets.QPushButton(parent=ListEditForm)
        self.delPushButton.setEnabled(False)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap("icons:edit_remove.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.delPushButton.setIcon(icon1)
        self.delPushButton.setObjectName("delPushButton")
        self.horizontalLayout.addWidget(self.delPushButton)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(ListEditForm)
        self.addPushButton.clicked.connect(ListEditForm.addEmptyItem) # type: ignore
        self.delPushButton.clicked.connect(ListEditForm.removeSelectedItem) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(ListEditForm)

    def retranslateUi(self, ListEditForm):
        _translate = QtCore.QCoreApplication.translate
        ListEditForm.setWindowTitle(_translate("ListEditForm", "Form"))
        self.label.setText(_translate("ListEditForm", "Label"))



# Form implementation generated from reading ui file 'DxSpots.ui'
#
# Created by: PyQt6 UI code generator 6.9.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_DxSpotsForm(object):
    def setupUi(self, DxSpotsForm):
        DxSpotsForm.setObjectName("DxSpotsForm")
        DxSpotsForm.resize(501, 476)
        self.verticalLayout = QtWidgets.QVBoxLayout(DxSpotsForm)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(parent=DxSpotsForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.nrSpotsSpinBox = QtWidgets.QSpinBox(parent=DxSpotsForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.nrSpotsSpinBox.sizePolicy().hasHeightForWidth())
        self.nrSpotsSpinBox.setSizePolicy(sizePolicy)
        self.nrSpotsSpinBox.setMinimum(10)
        self.nrSpotsSpinBox.setMaximum(200)
        self.nrSpotsSpinBox.setSingleStep(10)
        self.nrSpotsSpinBox.setProperty("value", 50)
        self.nrSpotsSpinBox.setObjectName("nrSpotsSpinBox")
        self.horizontalLayout.addWidget(self.nrSpotsSpinBox)
        spacerItem = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.label_2 = QtWidgets.QLabel(parent=DxSpotsForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout.addWidget(self.label_2)
        self.bandComboBox = QtWidgets.QComboBox(parent=DxSpotsForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bandComboBox.sizePolicy().hasHeightForWidth())
        self.bandComboBox.setSizePolicy(sizePolicy)
        self.bandComboBox.setObjectName("bandComboBox")
        self.horizontalLayout.addWidget(self.bandComboBox)
        self.line = QtWidgets.QFrame(parent=DxSpotsForm)
        self.line.setFrameShape(QtWidgets.QFrame.Shape.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout.addWidget(self.line)
        self.label_3 = QtWidgets.QLabel(parent=DxSpotsForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout.addWidget(self.label_3)
        self.spContComboBox = QtWidgets.QComboBox(parent=DxSpotsForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.spContComboBox.sizePolicy().hasHeightForWidth())
        self.spContComboBox.setSizePolicy(sizePolicy)
        self.spContComboBox.setObjectName("spContComboBox")
        self.horizontalLayout.addWidget(self.spContComboBox)
        spacerItem1 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.startPushButton = QtWidgets.QPushButton(parent=DxSpotsForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.startPushButton.sizePolicy().hasHeightForWidth())
        self.startPushButton.setSizePolicy(sizePolicy)
        self.startPushButton.setCheckable(True)
        self.startPushButton.setObjectName("startPushButton")
        self.horizontalLayout.addWidget(self.startPushButton)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.tableView = QtWidgets.QTableView(parent=DxSpotsForm)
        self.tableView.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)
        self.tableView.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.SingleSelection)
        self.tableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.tableView.setObjectName("tableView")
        self.tableView.verticalHeader().setVisible(False)
        self.verticalLayout.addWidget(self.tableView)

        self.retranslateUi(DxSpotsForm)
        self.startPushButton.toggled['bool'].connect(DxSpotsForm.control) # type: ignore
        self.bandComboBox.currentTextChanged['QString'].connect(DxSpotsForm.bandChanged) # type: ignore
        self.tableView.doubleClicked['QModelIndex'].connect(DxSpotsForm.selectSpot) # type: ignore
        self.spContComboBox.currentTextChanged['QString'].connect(DxSpotsForm.spContChanged) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(DxSpotsForm)

    def retranslateUi(self, DxSpotsForm):
        _translate = QtCore.QCoreApplication.translate
        DxSpotsForm.setWindowTitle(_translate("DxSpotsForm", "Form"))
        self.label.setText(_translate("DxSpotsForm", "Keep spots"))
        self.label_2.setText(_translate("DxSpotsForm", "Band"))
        self.label_3.setText(_translate("DxSpotsForm", "Sp. Cnt."))
        self.startPushButton.setText(_translate("DxSpotsForm", "Start"))



import re
import os.path
import logging
import platform
import tempfile
import subprocess

import requests
from adif_file import adi
from PyQt6 import QtCore
import xmltodict

from .Logger import Logger


class LoTWCommunicationException(Exception):
    pass


class LoTWRequestException(Exception):
    pass


class LoTWLoginException(Exception):
    pass


class LoTWNoRecordException(Exception):
    pass


class LoTWADIFFieldException(Exception):
    pass


mode_map = {}


def load_mode_map(log):
    if platform.system() == 'Windows':
        lotwcfg_path = os.path.expanduser('~/AppData/Roaming/TrustedQSL/config.xml')
    elif platform.system() == 'Linux':
        lotwcfg_path = os.path.expanduser('~/.tqsl/config.xml')
    else:
        log(logging.WARNING, f'System "{platform.system()}" is currently not supported to upload to LoTW')
        return

    if not os.path.isfile(lotwcfg_path):
        log(logging.WARNING, 'Not able to load TQSL configuration')
    else:
        try:
            with open(lotwcfg_path) as cfg_f:
                lotwcfg_xml = cfg_f.read()

            lotw_config = xmltodict.parse(lotwcfg_xml)

            for mode in lotw_config['tqslconfig']['adifmap']['adifmode']:
                if mode['@adif-mode'] not in mode_map:
                    if '@adif-submode' in mode:
                        continue
                    else:
                        mode_map[mode['@adif-mode']] = mode['@mode']

            log(logging.DEBUG, 'Loaded ADIF mode mapping')
        except Exception as exc:
            log(logging.ERROR, exc)


class LoTW:
    required_fields = ('QSO_DATE', 'TIME_ON', 'CALL', 'MODE', 'BAND')
    fields = required_fields + ('FREQ', 'QSLMSG', 'RST_SENT', 'MY_GRIDSQUARE', 'STATION_CALLSIGN')
    REGEX_RECORDS = re.compile(r'(?:.*?<[eE][oO][hH]>)?\n*(.*?<[eE][oO][rR]>).*?', re.DOTALL)

    def __init__(self, logger: Logger):
        self.log = logging.getLogger('LoTW')
        self.log.addHandler(logger)
        self.log.setLevel(logger.loglevel)
        self.logger = logger
        self.log.debug('Initialising...')

        if not mode_map:
            load_mode_map(self.log.log)

    def upload_log(self, station: str, doc: dict, password: str = '') -> bool:
        self._check_fields_(doc)

        # Export to tempfile
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_file = os.path.join(tmp_dir, 'DragonLog_Export.adi')
            adi.dump(tmp_file, doc, 'ADIF Export by DragonLog')

            tqsl_startupinfo = None
            if platform.system() == 'Windows':
                tqsl_path = 'C:/Program Files (x86)/TrustedQSL/tqsl.exe'
                tqsl_startupinfo = subprocess.STARTUPINFO()
                tqsl_startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            else:
                tqsl_path = 'tqsl'

            cmd = [tqsl_path,
                   '-d',  # Supress date range dialog
                   '-u',  # Directly upload
                   '-a', 'all',  # Sign all including already sent
                   '-f', 'ignore',  # Ignore QTH information
                   '-x',  # Batch mode
                   '-l', station,
                   tmp_file]
            if password:
                cmd.append('-p')
                cmd.append(password)

            res = subprocess.run(cmd, capture_output=True, startupinfo=tqsl_startupinfo)
            match res.returncode:
                case 0:
                    self.log.debug('TQSL exited with success')
                    return True
                case 1:  # Will it be possible?
                    self.log.debug('TQSL canceled by user')
                case 2:
                    self.log.warning('Log rejected by LoTW')
                    raise LoTWRequestException
                case 3 | 4 | 5:
                    self.log.error(f'Local or server error: {res.returncode}')
                case 6 | 7:  # Should not occur on tempfile
                    self.log.error(f'Error for read/write on input/output file: {res.returncode}')
                case 8:  # Should never occur
                    self.log.warning('TQSL no QSOs were processed')
                case 9:
                    self.log.warning('TQSL some QSOs were already uploaded')
                    return True
                case 10:  # Should not occur due to tested syntax
                    self.log.error('TQSL command syntax error')
                case 11:
                    self.log.warning('LoTW Connection error or network unreachable')
                    raise LoTWCommunicationException()

        return False

    def _check_fields_(self, doc: dict):
        for i, record in enumerate(doc['RECORDS']):
            for field in self.required_fields:
                if field not in record:
                    raise LoTWADIFFieldException(f'{field} in record #{i + 1}')

    def check_inbox(self, username: str, password: str, record: dict) -> bool:
        self._check_fields_({'RECORDS': [record]})
        self.log.debug(f"Checking inbox for {record['CALL']} on {record['QSO_DATE']}")

        if not username or not password:
            raise LoTWLoginException('Missing username or password')

        qso_dt = QtCore.QDateTime.fromString(
            f"{record['QSO_DATE'][:4]}-{record['QSO_DATE'][4:6]}-{record['QSO_DATE'][6:8]} "
            f"{record['TIME_ON'][:2]}:{record['TIME_ON'][2:4]}",
            'yyyy-MM-dd hh:mm')
        start_dt = qso_dt.addSecs(-300)
        end_dt = qso_dt.addSecs(300)

        self.log.debug(f"Filter QSOs between {start_dt.toString('yyyy-MM-dd hh:mm')} and "
                       f"{end_dt.toString('yyyy-MM-dd hh:mm')}")

        params = {
            'login': username,
            'password': password,
            'qso_query': '1',
            'qso_callsign': record['CALL'],
            'qso_startdate': start_dt.date().toString('yyyy-MM-dd'),
            'qso_starttime': start_dt.time().toString('hh:mm'),
            'qso_enddate': end_dt.date().toString('yyyy-MM-dd'),
            'qso_endtime': end_dt.time().toString('hh:mm'),
            'qso_band': record['BAND'],
            'qso_mode': mode_map[record['MODE']] if record['MODE'] in mode_map else record['MODE'],
        }

        r = requests.get('https://lotw.arrl.org/lotwuser/lotwreport.adi', params=params)
        if r.status_code == 200:
            if r.text.strip().endswith('<APP_LoTW_EOF>'):
                records = re.findall(self.REGEX_RECORDS, r.text)
                if len(records) > 1:
                    raise LoTWRequestException('Too much search results')
                elif len(records) == 0:
                    raise LoTWNoRecordException('No search results')

                try:
                    doc = adi.loads(records[0].strip())

                    if doc['RECORDS'][0]['QSL_RCVD'] == 'Y':
                        return True
                    else:
                        return False
                except adi.TagDefinitionException as exc:
                    raise LoTWRequestException(exc)
            else:
                raise LoTWLoginException()
        else:
            raise LoTWCommunicationException(f'LoTW error: HTTP-Error {r.status_code}')



import os.path
import sys
import time
import logging

from PyQt6 import QtCore, QtWidgets, QtGui


class BaseLogger(logging.Handler):
    def __init__(self, loggername: str | None = None, loglevel: str = 'INFO'):
        super().__init__()

        self.__loglevel__ = loglevel
        logging.Formatter.converter = time.gmtime
        self.setFormatter(logging.Formatter('%(asctime)s %(levelname)-8s: %(name)s - %(message)s'))
        self.__log__ = logging.getLogger(loggername)
        self.__log__.setLevel(self.__loglevel__)
        self.__log__.addHandler(self)

    @property
    def loglevel(self) -> str:
        return self.__loglevel__

    def emit(self, record):
        log_msg = self.format(record)

        if record.levelno >= 30:  # warning, error, critical
            print(log_msg, file=sys.stderr)
        else:  # info, debug
            print(log_msg)

    def debug(self, message):
        self.__log__.debug(message)

    def info(self, message):
        self.__log__.info(message)

    def warning(self, message):
        self.__log__.warning(message)

    def error(self, message):
        self.__log__.error(message)

    def critical(self, message):
        self.__log__.critical(message)

    def exception(self, message):
        self.__log__.exception(message)


class Logger(BaseLogger):
    def __init__(self, log_widget: QtWidgets.QTextEdit, settings: QtCore.QSettings):
        super().__init__('DragonLog', settings.value('ui/log_level', 'INFO'))
        self.log_widget = log_widget

        self.__log_file__ = None
        if int(settings.value('ui/log_file', 0)):
            try:
                self.__log_file__ = open(os.path.expanduser('~/DragonLog.log'), 'a', buffering=1, encoding='utf8')
            except Exception as exc:
                self.error('Log file could not be opened')
                self.exception(exc)

    @property
    def loglevel(self) -> str:
        return self.__loglevel__

    def emit(self, record):
        super().emit(record)

        log_msg = self.format(record)

        if record.levelno >= 40:  # Error/Critical
            self.log_widget.setTextColor(QtGui.QColor('red'))
        elif record.levelno >= 30:  # Warning
            self.log_widget.setTextColor(QtGui.QColor('orange'))
        elif record.levelno < 20:  # Debug
            self.log_widget.setTextColor(QtGui.QColor('grey'))
        else:  # Info >= 20
            self.log_widget.setTextColor(QtGui.QColor('black'))

        if self.__log_file__:
            self.__log_file__.write(log_msg + '\n')

        self.log_widget.append(log_msg)
        self.log_widget.verticalScrollBar().setValue(self.log_widget.verticalScrollBar().maximum())
        self.log_widget.repaint()



# DragonLog (c) 2023-2025 by Andreas Schawo is licensed under CC BY-SA 4.0.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/

import os
import sys
import typing
import logging
import platform

from PyQt6 import QtWidgets, QtCore
import keyring

from . import DragonLog_Settings_ui
from .Logger import Logger
from .RegEx import REGEX_CALL, check_format
from .CallBook import CallBookType
from .ListEdit import ListEdit
from .SelectWidget import SelectWidget
from . import ColorPalettes
from .RigControl import RigControl, RigctldExecutionException, NoExecutableFoundException

# Fix problems with importing win32 in frozen executable
if getattr(sys, 'frozen', False):
    # noinspection PyUnresolvedReferences
    import win32timezone
    from keyring.backends import Windows

    keyring.set_keyring(Windows.WinVaultKeyring())


class Settings(QtWidgets.QDialog, DragonLog_Settings_ui.Ui_Dialog):
    callbookChanged = QtCore.pyqtSignal(str)
    settingsStored = QtCore.pyqtSignal()
    ctyDataChanged = QtCore.pyqtSignal(str)

    def __init__(self, parent, settings: QtCore.QSettings,
                 bands: typing.Iterable, modes: typing.Iterable, cols: typing.Iterable,
                 logger: Logger, rigctl: RigControl):
        super().__init__(parent)
        self.setupUi(self)

        self.log = logging.getLogger('Settings')
        self.log.addHandler(logger)
        self.log.setLevel(logger.loglevel)
        self.logger = logger
        self.log.debug('Initialising...')

        self.qthsListEdit = ListEdit(self.listingsWidget, self.tr('QTHs & locators'))
        self.qthsListEdit.setToolTip(self.tr('Type in as: "City (Locator)", i.e. "Koblenz (JO30ti)"'))
        self.listingsWidget.layout().addWidget(self.qthsListEdit, 0, 0, 1, 2)
        self.qthsListEdit.listChanged.connect(self.refreshQTHsComboBox)
        self.rigsListEdit = ListEdit(self.listingsWidget, self.tr('Radios'))
        self.listingsWidget.layout().addWidget(self.rigsListEdit, 1, 0, 1, 1)
        self.rigsListEdit.listChanged.connect(self.refreshRigsComboBox)
        self.antsListEdit = ListEdit(self.listingsWidget, self.tr('Antennas'))
        self.listingsWidget.layout().addWidget(self.antsListEdit, 1, 1, 1, 1)
        self.antsListEdit.listChanged.connect(self.refreshAntsComboBox)

        self.bandsSelectWidget = SelectWidget(self.bandsGroupBox, bands,
                                              self.tr('Show bands'), self.tr('Hide bands'))
        self.bandsGroupBox.layout().addWidget(self.bandsSelectWidget)
        self.modesSelectWidget = SelectWidget(self.modesGroupBox, modes,
                                              self.tr('Show modes'), self.tr('Hide modes'))
        self.modesGroupBox.layout().addWidget(self.modesSelectWidget)
        self.columnsSelectWidget = SelectWidget(self.columnsGroupBox, cols,
                                                self.tr('Show columns'), self.tr('Hide columns'))
        self.columnsGroupBox.layout().addWidget(self.columnsSelectWidget)

        self.settings = settings

        self.rigctl = rigctl

        if platform.system() == 'Windows':
            if self.settings.value('cat/rigctldPath', None):
                if self.rigctl.__is_exe__(self.settings.value('cat/rigctldPath')):
                    self.hamlibPathLineEdit.setText(self.settings.value('cat/rigctldPath'))
                else:
                    self.settings.setValue('cat/rigctldPath', '')
                    self.hamlibPathLineEdit.setText('')
        else:
            self.rigctldPathLabel.setVisible(False)
            self.hamlibPathLineEdit.setVisible(False)
            self.hamlibPathToolButton.setVisible(False)
            self.checkHamlibLabel.setVisible(False)

        self.refreshMfrComboBox()

        self.columns = cols
        self.sortComboBox.insertItems(0, cols)

        self.callbooks = dict([(cbt.value, cbt.name) for cbt in set(CallBookType)])
        self.callbookComboBox.insertItems(0, self.callbooks.keys())

    def refreshMfrComboBox(self):
        self.manufacturerComboBox.clear()
        self.manufacturerComboBox.insertItems(0, sorted(self.rigctl.availableManufacturers))
        if self.settings.value('cat/rigMfr', None):
            self.manufacturerComboBox.setCurrentText(self.settings.value('cat/rigMfr'))
        else:
            self.manufacturerComboBox.setCurrentIndex(0)

    def refreshQTHsComboBox(self):
        self.qthComboBox.clear()
        self.qthComboBox.insertItems(0, self.qthsListEdit.items())
        self.qthComboBox.setCurrentText(self.settings.value('station/qth_loc', ''))

    def refreshRigsComboBox(self):
        self.radioComboBox.clear()
        self.radioComboBox.insertItems(0, self.rigsListEdit.items())
        self.radioComboBox.setCurrentText(self.settings.value('station/radio', ''))

    def refreshAntsComboBox(self):
        self.antennaComboBox.clear()
        self.antennaComboBox.insertItems(0, self.antsListEdit.items())
        self.antennaComboBox.setCurrentText(self.settings.value('station/antenna', ''))

    @staticmethod
    def __is_exe__(path):
        return os.path.isfile(path) and os.access(path, os.X_OK)

    def chooseHamlibPath(self):
        rigctld_path = QtWidgets.QFileDialog.getOpenFileName(
            self,
            self.tr('Choose hamlib rigctld executable'),
            self.settings.value('cat/rigctldPath', None),
            'rigctld.exe (rigctld.exe)'
        )

        if self.__is_exe__(rigctld_path[0]):
            try:
                self.rigctl.init_hamlib(rigctld_path[0])
                self.settings.setValue('cat/rigctldPath', rigctld_path[0])
                self.checkHamlibLabel.setText('')
                self.hamlibPathLineEdit.setText(rigctld_path[0])
                self.refreshMfrComboBox()
            except RigctldExecutionException as exc:
                self.settings.setValue('cat/rigctldPath', '')
                self.checkHamlibLabel.setText(self.tr('Error executing rigctld'))
                self.hamlibPathLineEdit.setText('')
            except NoExecutableFoundException as exc:
                self.settings.setValue('cat/rigctldPath', '')
                self.checkHamlibLabel.setText(self.tr('rigctld is not available'))
        else:
            self.checkHamlibLabel.setText(self.tr('Selected file is not the executable'))

    def mfrChanged(self, mfr):
        self.modelComboBox.clear()

        if mfr in self.rigctl.availableManufacturers:
            self.modelComboBox.insertItems(0, sorted(self.rigctl.availableRigs(mfr)))

        if self.settings.value('cat/rigModel', None):
            self.modelComboBox.setCurrentText(self.settings.value('cat/rigModel'))

    def callSignChanged(self, txt):
        if not txt:
            self.callsignLineEdit.setPalette(ColorPalettes.PaletteEmpty)
        elif check_format(REGEX_CALL, txt):
            self.callsignLineEdit.setPalette(ColorPalettes.PaletteOk)
        else:
            self.callsignLineEdit.setPalette(ColorPalettes.PaletteFaulty)

    def callSignCBChanged(self, txt):
        if not txt:
            self.callsignCBLineEdit.setPalette(ColorPalettes.PaletteEmpty)
        else:
            self.callsignCBLineEdit.setPalette(ColorPalettes.PaletteOk)

    def callbookSelected(self, service: str):
        """Slot (i.e. if callbookComboBox is changed)"""
        self.callbookUserLineEdit.setText(self.settings.value(f'callbook/{self.callbooks[service]}_user', ''))

    def chooseCtyDataFile(self):
        cty_path = QtWidgets.QFileDialog.getOpenFileName(
            self,
            self.tr('Choose Country data file'),
            os.path.split(self.settings.value('dx_spots/cty_data', ''))[0],
            'Country data (*.csv)'
        )

        if os.path.isfile(cty_path[0]):
            self.ctyPathLineEdit.setText(cty_path[0])
            self.settings.setValue('dx_spots/cty_data', cty_path[0])
            self.ctyDataChanged.emit(cty_path[0])

    def exec(self):
        self.log.info('Loading settings...')
        self.catInterfaceLineEdit.setText(self.settings.value('cat/interface', ''))
        self.catBaudComboBox.setCurrentText(self.settings.value('cat/baud', ''))

        self.qthsListEdit.clear()
        self.qthsListEdit.setItems(self.settings.value('listings/qths'))
        self.rigsListEdit.clear()
        self.rigsListEdit.setItems(self.settings.value('listings/rigs'))
        self.antsListEdit.clear()
        self.antsListEdit.setItems(self.settings.value('listings/antennas'))

        self.callsignLineEdit.setText(self.settings.value('station/callSign', '').upper())
        self.nameLineEdit.setText(self.settings.value('station/name', ''))
        self.qthComboBox.setCurrentText(self.settings.value('station/qth_loc', ''))
        self.radioComboBox.setCurrentText(self.settings.value('station/radio', ''))
        self.antennaComboBox.setCurrentText(self.settings.value('station/antenna', ''))

        self.callsignCBLineEdit.setText(self.settings.value('station_cb/callSign', ''))
        self.radioCBLineEdit.setText(self.settings.value('station_cb/radio', ''))
        self.antennaCBLineEdit.setText(self.settings.value('station_cb/antenna', ''))
        self.cbDefaultCheckBox.setChecked(bool(int(self.settings.value('station_cb/cb_by_default', 0))))
        self.expCBQSOsCheckBox.setChecked(bool(int(self.settings.value('station_cb/cb_exp_adif', 0))))

        self.fontComboBox.setCurrentText(self.settings.value('ui/font', 'Inter SlashedZero'))
        self.fontSizeSpinBox.setValue(int(self.settings.value('ui/font_size', 9)))
        self.qfFontSizeSpinBox.setValue(int(self.settings.value('ui/form_font_size', 10)))
        self.ccFontSizeSpinBox.setValue(int(self.settings.value('ui/cc_font_size', 12)))
        self.sortComboBox.setCurrentText(self.settings.value('ui/sort_col', self.tr('Date/Time start')))
        sort_order = self.settings.value('ui/sort_order', 'ASC')
        self.sortAscRadioButton.setChecked(sort_order == 'ASC')
        self.sortDscRadioButton.setChecked(sort_order == 'DSC')
        self.recentQSOsComboBox.setCurrentText(self.settings.value('ui/recent_qsos', self.tr('Show all')))

        try:
            self.columnsSelectWidget.indexesDisabled = [int(c) for c in self.settings.value('ui/hidden_cols',
                                                                                            '1').split(',')]
        except ValueError:
            self.logger.exception(f'Reading "ui/hidden_cols": {self.settings.value("ui/hidden_cols", "")}')

        self.bandsSelectWidget.itemsEnabled = self.settings.value('ui/show_bands', self.bandsSelectWidget.items)
        self.modesSelectWidget.itemsEnabled = self.settings.value('ui/show_modes', self.modesSelectWidget.items)

        self.checkUpdatesCheckBox.setChecked(bool(int(self.settings.value('ui/check_updates', 1))))
        self.logLevelComboBox.setCurrentText(str(self.settings.value('ui/log_level', 'Info')).capitalize())
        self.logToFileCheckBox.setChecked(bool(int(self.settings.value('ui/log_file', 0))))

        dx_call = self.settings.value('dx_spots/call', '')
        self.dxCallLineEdit.setText(dx_call if dx_call else self.settings.value('station/callSign', ''))
        dx_addr = self.settings.value('dx_spots/address', '')
        self.dxAddrLineEdit.setText(dx_addr if dx_addr else 'hamqth.com')
        dx_port = int(self.settings.value('dx_spots/port', 7300)) if dx_addr else 7300
        self.dxPortSpinBox.setValue(dx_port if 0 < dx_port <= self.dxPortSpinBox.maximum() else 7300)
        self.ctyPathLineEdit.setText(self.settings.value('dx_spots/cty_data', ''))

        self.expOwnNotesADIFCheckBox.setChecked(bool(int(self.settings.value('imp_exp/own_notes_adif', 0))))
        self.expRecentOnlyCheckBox.setChecked(bool(int(self.settings.value('imp_exp/only_recent', 0))))
        self.csvDelimiterComboBox.setCurrentText(self.csv_delimiter)
        self.useCfgIDImportCheckBox.setChecked(bool(int(self.settings.value('imp_exp/use_id_adif', 0))))
        self.useCfgStationImportCheckBox.setChecked(bool(int(self.settings.value('imp_exp/use_station_adif', 0))))
        self.useCfgIDWatchCheckBox.setChecked(bool(int(self.settings.value('imp_exp/use_id_watch', 0))))
        self.useCfgStationWatchCheckBox.setChecked(bool(int(self.settings.value('imp_exp/use_station_watch', 0))))
        self.useCfgStationCCCheckBox.setChecked(bool(int(self.settings.value('imp_exp/use_station_hamcc', 0))))

        self.callbookComboBox.setCurrentText(self.callbook_dname)
        self.callbookUserLineEdit.setText(self.settings.value(f'callbook/{self.callbook_id}_user', ''))

        self.eqslUserLineEdit.setText(self.settings.value('eqsl/username', ''))
        self.lotwUserLineEdit.setText(self.settings.value('lotw/username', ''))
        self.lotwCertPwdCheckBox.setChecked(bool(int(self.settings.value('lotw/cert_needs_pwd', 0))))

        return super().exec()

    @property
    def csv_delimiter(self) -> str:
        return self.settings.value('imp_exp/csv_delimiter', ';' if self.locale().decimalPoint() == ',' else ',')

    @property
    def callbook_id(self) -> str:
        """The selected callbooks id"""
        return self.settings.value('callbook/service', 'HamQTH')

    @property
    def callbook_dname(self) -> str:
        """Returns the selected callbooks descriptive name"""
        return CallBookType[self.settings.value('callbook/service', 'HamQTH')].value

    def callbookPassword(self, callbook: CallBookType) -> str:
        """Get the password for the callbook
        :param callbook: the callbook service to get the password for
        :return: the password"""

        return keyring.get_password(callbook.value,
                                    self.settings.value(f'callbook/{callbook.name}_user', ''))

    def eqslPassword(self) -> str:
        """The password for the eQSL service"""
        return keyring.get_password('eqsl.cc',
                                    self.settings.value('eqsl/username', ''))

    def lotwPassword(self) -> str:
        """The password for the LoTW online service"""
        return keyring.get_password('lotw.arrl.org',
                                    self.settings.value('lotw/username', ''))

    def accept(self):
        self.log.info('Saving Settings...')
        self.settings.setValue('cat/interface', self.catInterfaceLineEdit.text())
        self.settings.setValue('cat/baud', self.catBaudComboBox.currentText())
        self.settings.setValue('cat/rigMfr', self.manufacturerComboBox.currentText())
        self.settings.setValue('cat/rigModel', self.modelComboBox.currentText())

        self.settings.setValue('station/callSign', self.callsignLineEdit.text().upper())
        self.settings.setValue('station/name', self.nameLineEdit.text())
        self.settings.setValue('station/qth_loc', self.qthComboBox.currentText())
        self.settings.setValue('station/radio', self.radioComboBox.currentText())
        self.settings.setValue('station/antenna', self.antennaComboBox.currentText())

        self.settings.setValue('listings/qths', self.qthsListEdit.items())
        self.settings.setValue('listings/rigs', self.rigsListEdit.items())
        self.settings.setValue('listings/antennas', self.antsListEdit.items())

        self.settings.setValue('station_cb/callSign', self.callsignCBLineEdit.text())
        self.settings.setValue('station_cb/radio', self.radioCBLineEdit.text())
        self.settings.setValue('station_cb/antenna', self.antennaCBLineEdit.text())
        self.settings.setValue('station_cb/cb_by_default', int(self.cbDefaultCheckBox.isChecked()))
        self.settings.setValue('station_cb/cb_exp_adif', int(self.expCBQSOsCheckBox.isChecked()))

        self.settings.setValue('ui/font', self.fontComboBox.currentText())
        self.settings.setValue('ui/font_size', self.fontSizeSpinBox.value())
        self.settings.setValue('ui/form_font_size', self.qfFontSizeSpinBox.value())
        self.settings.setValue('ui/cc_font_size', self.ccFontSizeSpinBox.value())
        self.settings.setValue('ui/sort_col', self.sortComboBox.currentText())
        self.settings.setValue('ui/sort_order', 'ASC' if self.sortAscRadioButton.isChecked() else 'DSC')
        self.settings.setValue('ui/recent_qsos', self.recentQSOsComboBox.currentText())
        self.settings.setValue('ui/hidden_cols',
                               ','.join([str(c) for c in self.columnsSelectWidget.indexesDisabled]))
        self.settings.setValue('ui/show_bands', self.bandsSelectWidget.itemsEnabled)
        self.settings.setValue('ui/show_modes', self.modesSelectWidget.itemsEnabled)
        self.settings.setValue('ui/check_updates', int(self.checkUpdatesCheckBox.isChecked()))

        self.settings.setValue('ui/log_level', self.logLevelComboBox.currentText().upper())
        self.settings.setValue('ui/log_file', int(self.logToFileCheckBox.isChecked()))

        self.settings.setValue('dx_spots/call', self.dxCallLineEdit.text())
        self.settings.setValue('dx_spots/address', self.dxAddrLineEdit.text())
        self.settings.setValue('dx_spots/port', self.dxPortSpinBox.value())
        self.settings.setValue('dx_spots/cty_data', self.ctyPathLineEdit.text())

        self.settings.setValue('imp_exp/own_notes_adif', int(self.expOwnNotesADIFCheckBox.isChecked()))
        self.settings.setValue('imp_exp/only_recent', int(self.expRecentOnlyCheckBox.isChecked()))
        self.settings.setValue('imp_exp/csv_delimiter', self.csvDelimiterComboBox.currentText())
        self.settings.setValue('imp_exp/use_id_adif', int(self.useCfgIDImportCheckBox.isChecked()))
        self.settings.setValue('imp_exp/use_station_adif', int(self.useCfgStationImportCheckBox.isChecked()))
        self.settings.setValue('imp_exp/use_id_watch', int(self.useCfgIDWatchCheckBox.isChecked()))
        self.settings.setValue('imp_exp/use_station_watch', int(self.useCfgStationWatchCheckBox.isChecked()))
        self.settings.setValue('imp_exp/use_station_hamcc', int(self.useCfgStationCCCheckBox.isChecked()))

        self.settings.setValue('callbook/service', self.callbooks[self.callbookComboBox.currentText()])
        self.settings.setValue(f'callbook/{self.callbook_id}_user', self.callbookUserLineEdit.text())
        if self.callbookUserLineEdit.text() and self.callbookPasswdLineEdit.text():
            keyring.set_password(self.callbookComboBox.currentText(),
                                 self.callbookUserLineEdit.text(),
                                 self.callbookPasswdLineEdit.text())
        self.callbookPasswdLineEdit.clear()
        self.callbookChanged.emit(self.callbooks[self.callbookComboBox.currentText()])

        self.settings.setValue('eqsl/username', self.eqslUserLineEdit.text())
        if self.eqslUserLineEdit.text() and self.eqslPasswdLineEdit.text():
            keyring.set_password('eqsl.cc',
                                 self.eqslUserLineEdit.text(),
                                 self.eqslPasswdLineEdit.text())
        self.eqslPasswdLineEdit.clear()

        self.settings.setValue('lotw/username', self.lotwUserLineEdit.text())
        if self.lotwUserLineEdit.text() and self.lotwPasswdLineEdit.text():
            keyring.set_password('lotw.arrl.org',
                                 self.lotwUserLineEdit.text(),
                                 self.lotwPasswdLineEdit.text())
        self.lotwPasswdLineEdit.clear()
        self.settings.value('lotw/cert_needs_pwd', int(self.lotwCertPwdCheckBox.isChecked()))

        self.settingsStored.emit()
        super().accept()



# DragonLog (c) 2025 by Andreas Schawo is licensed under CC BY-SA 4.0.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/

from .base import ContestLog, ExchangeData, ContestLogCBR, ContestLogEDI
from .contests_darc import DARCUKWContest, DARCUKWFruehlingsContest, DARCOsterContest, DARCUKWSommerFDContest
from .contests_darc_rlp import RLPFALZAWLog, RLPFALZABKWLog, RLPFALZABUKWLog, K32KurzUKWLog
from .contests_darc_others import L33EinsteigerContest
from .contests_others import IARUHFWorldChampionshipContest

CONTESTS: dict[str, type[ContestLog]] = {
    'DARC-UKW': DARCUKWContest,
    'DARC-UKW-FRUEHLING': DARCUKWFruehlingsContest,
    'DARC-UKW-SOMMERFD': DARCUKWSommerFDContest,
    'DARC-KW-OSTERN': DARCOsterContest,
    'RL-PFALZ-AW': RLPFALZAWLog,
    'RL-PFALZ-AB.UKW': RLPFALZABUKWLog,
    'RL-PFALZ-AB.KW': RLPFALZABKWLog,
    'K32-KURZ-UKW': K32KurzUKWLog,
    'L33-EINSTEIGER': L33EinsteigerContest,
    'IARU-HF': IARUHFWorldChampionshipContest,
    # Insert other contests above
    'KW-UNIVERSAL': ContestLogCBR,
    'UKW-UNIVERSAL': ContestLogEDI,
}

CONTEST_NAMES = dict(zip(CONTESTS.keys(), [c.contest_name for c in CONTESTS.values()]))
CONTEST_IDS = dict(zip([c.contest_name for c in CONTESTS.values()], CONTESTS.keys()))


def build_contest_list() -> str:
    """Build a contest list as Markdown"""

    text = '''
Available Contests
==================

The table shows all available contests with the last date the contest definition was updated.

The *Year* column shows the year the contest definition is targeted to. 
If it does not show the current year you should check for a program update.

The *Internal ID* is the ID which is imported or exported in ADIF format. 

The *Exch format* is the format you must use to type in the received exchange. 
If a separator is you can also use a blank or underscore instead of a comma.
The sent exchange is handled automatically.

| Contest name | Internal ID | Year | Updated | Exch format |
|--------------|-------------|------|---------|-------------|
'''

    for c in CONTESTS:
        cntst: type[ContestLog] = CONTESTS[c]
        text += f'| {cntst.contest_name} | {c} | ***{cntst.contest_year}*** | {cntst.contest_update} | {cntst.contest_exch_fmt} |\n'

    return text



# DragonLog (c) 2025 by Andreas Schawo is licensed under CC BY-SA 4.0.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/

from .base import ContestLogEDI, CategoryMode, CategoryBand


class L33EinsteigerContest(ContestLogEDI):
    contest_name = 'L33 Einsteiger-Contest'
    contest_year = '2025'
    contest_update = '2025-04-06'

    @classmethod
    def valid_modes(cls) -> tuple[CategoryMode, ...]:
        return CategoryMode.FM,

    @classmethod
    def valid_bands(cls) -> tuple[CategoryBand, ...]:
        return CategoryBand.B_2M,

    @classmethod
    def descr_specific(cls) -> str:
        return 'DOK'



# DragonLog (c) 2023-2025 by Andreas Schawo is licensed under CC BY-SA 4.0.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/

import logging

from PyQt6 import QtWidgets, QtCore

from . import DragonLog_QSOForm_ui
from .Logger import Logger
from .DragonLog_Settings import Settings
from .RegEx import REGEX_CALL, REGEX_RSTFIELD, REGEX_LOCATOR, REGEX_TIME, check_format, check_call, check_qth
from .CallBook import (HamQTHCallBook, QRZCQCallBook, QRZCallBook, CallBookData,
                       SessionExpiredException, LoginException, CallsignNotFoundException)
from . import ColorPalettes
from .contest import CONTESTS, CONTEST_IDS, CONTEST_NAMES, ContestLog, ExchangeData
from .distance import distance
from .cty import Country
from .local_callbook import LocalCallbook


class QSOForm(QtWidgets.QDialog, DragonLog_QSOForm_ui.Ui_QSOForm):
    def __init__(self, parent, dragonlog, bands: dict, modes: dict, prop: dict, settings: QtCore.QSettings,
                 settings_form: Settings, cb_channels: dict, logger: Logger, local_cb: LocalCallbook):
        super().__init__(parent)
        self.dragonlog = dragonlog
        self.setupUi(self)

        self.log = logging.getLogger('QSOForm')
        self.log.addHandler(logger)
        self.log.setLevel(logger.loglevel)
        self.logger = logger
        self.log.debug('Initialising...')

        self.__local_cb__ = local_cb

        self.lastpos = None
        self.bands = bands
        self.modes = modes
        self.prop = prop
        self.prop_trans = dict(zip(prop.values(), prop.keys()))
        self.settings = settings
        self.settings_form = settings_form
        self.qso_id = None

        self.__old_values__ = {}

        self.cb_channels = cb_channels
        self.channelComboBox.insertItems(0, ['-'] + list(cb_channels.keys()))

        self.refreshBands()

        self.propComboBox.insertItems(0, self.prop.values())

        self.stationChanged(True)
        self.identityChanged(True)

        self.__change_mode__ = False

        self.timeTimer = QtCore.QTimer(self)
        self.timeTimer.timeout.connect(self.refreshTime)

        self.worked_dialog: QtWidgets.QListWidget | None = None
        self._create_worked_dlg_()

        self.callbook_hamqth = HamQTHCallBook(self.logger,
                                              f'{self.dragonlog.programName}-{self.dragonlog.programVersion}',
                                              )
        self.callbook_qrzcq = QRZCQCallBook(self.logger,
                                            f'{self.dragonlog.programName}-{self.dragonlog.programVersion}',
                                            )
        self.callbook_qrz = QRZCallBook(self.logger,
                                        f'{self.dragonlog.programName}-{self.dragonlog.programVersion}',
                                        )

        view_only_widgets = (
            self.qslAccBureauCheckBox,
            self.qslAccDirectCheckBox,
            self.qslAcceQSLCheckBox,
            self.qslAccLoTWCheckBox,
        )

        for w in view_only_widgets:
            w.setAttribute(QtCore.Qt.WidgetAttribute.WA_TransparentForMouseEvents)
            w.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

        self.refreshQTHList()
        self.refreshRadioList()
        self.refreshAntennaList()

        self.contestComboBox.insertItem(0, '')
        self.contestComboBox.insertItems(1, CONTEST_IDS.keys())

        self.clear()

    def refreshOwnData(self):
        if self.identityGroupBox.isChecked():
            self.ownNameLineEdit.setText(self.settings.value('station/name', ''))

            if self.bandComboBox.currentText() == '11m':
                self.ownCallSignLineEdit.setText(self.settings.value('station_cb/callSign', ''))
            else:
                self.ownCallSignLineEdit.setText(self.settings.value('station/callSign', ''))

        if self.stationGroupBox.isChecked():
            self.ownQTHComboBox.setCurrentText(self.settings.value('station/qth_loc', ''))

            if self.bandComboBox.currentText() == '11m':
                self.radioComboBox.setCurrentText(self.settings.value('station_cb/radio', ''))
                self.antennaComboBox.setCurrentText(self.settings.value('station_cb/antenna', ''))
            else:
                self.radioComboBox.setCurrentText(self.settings.value('station/radio', ''))
                self.antennaComboBox.setCurrentText(self.settings.value('station/antenna', ''))

    def setBand(self, band: str):
        self.bandComboBox.setCurrentText(band)

    def setFrequency(self, freq: float):
        self.freqDoubleSpinBox.setValue(freq)

    def setMode(self, mode: str):
        self.modeComboBox.setCurrentText(mode)

    def setSubmode(self, submode: str):
        self.submodeComboBox.setCurrentText(submode)

    def setPower(self, power: int):
        self.powerSpinBox.setValue(power)

    def refreshBands(self):
        self.bandComboBox.clear()
        self.bandComboBox.insertItems(0, self.settings.value('ui/show_bands', self.bands.keys()))
        self.bandComboBox.setCurrentIndex(0)

    def refreshQTHList(self):
        self.ownQTHComboBox.clear()
        if self.settings.value('listings/qths'):
            self.ownQTHComboBox.insertItems(0, self.settings.value('listings/qths'))

    def refreshRadioList(self):
        self.radioComboBox.clear()
        if self.settings.value('listings/rigs'):
            self.radioComboBox.insertItems(0, self.settings.value('listings/rigs'))

    def refreshAntennaList(self):
        self.antennaComboBox.clear()
        if self.settings.value('listings/antennas'):
            self.antennaComboBox.insertItems(0, self.settings.value('listings/antennas'))

    def startTimers(self, start: bool):
        if start:
            self.timeTimer.start(1000)
        else:
            self.timeTimer.stop()

    def _create_worked_dlg_(self):
        self.worked_dialog = QtWidgets.QListWidget(self)
        self.worked_dialog.setMinimumHeight(100)
        self.worked_dialog.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.NoSelection)
        self.worked_dialog.setSortingEnabled(True)
        self.worked_dialog.hide()

    def _callbook_lookup_(self):
        call = self.callSignLineEdit.text().strip()
        if not self.__local_cb__ or not call:
            return

        contest = CONTEST_IDS.get(self.contestComboBox.currentText(), None)
        if contest and self.__local_cb__.history_entries[0]:
            ch_data = self.__local_cb__.lookup_history(contest, call, True)
            if ch_data:
                if ch_data[2] and not self.rcvdDataLineEdit.text():
                    exc_data = ExchangeData(locator=ch_data[2].locator,
                                            power=ch_data[2].power_class,
                                            darc_dok=ch_data[2].darc_dok,
                                            itu_zone=ch_data[2].itu_zone)
                    contest = CONTESTS.get(contest, None)
                    self.rcvdDataLineEdit.setText(contest.prepare_exchange(exc_data))

                if ch_data[3]:
                    if ch_data[3].name and not self.nameLineEdit.text():
                        self.nameLineEdit.setText(ch_data[3].name)
                    if ch_data[3].qth and not self.QTHLineEdit.text():
                        self.QTHLineEdit.setText(ch_data[3].qth)
                    if ch_data[3].locator and not self.locatorLineEdit.text():
                        self.locatorLineEdit.setText(ch_data[3].locator)
        else:
            lcd = self.__local_cb__.lookup(call, True)
            if lcd:
                if not self.nameLineEdit.text().strip() and lcd[1].name:
                    self.nameLineEdit.setText(lcd[1].name)
                if not self.QTHLineEdit.text().strip() and lcd[1].qth:
                    self.QTHLineEdit.setText(lcd[1].qth)
                if not self.locatorLineEdit.text().strip() and lcd[1].locator:
                    self.locatorLineEdit.setText(lcd[1].locator)
                    self.locatorChanged(lcd[1].locator)

    def callEditingFinished(self):
        self.setWorkedBefore()
        self._callbook_lookup_()

    def setWorkedBefore(self, worked: list = None):
        self.worked_dialog.clear()
        if worked:
            self.worked_dialog.addItems(worked)
            call_edit_pos = self.callSignLineEdit.pos()
            call_edit_pos.setX(call_edit_pos.x() + 15)
            call_edit_pos.setY(call_edit_pos.y() + self.callSignLineEdit.height() + 40)
            self.worked_dialog.move(call_edit_pos)
            self.worked_dialog.show()
        else:
            self.worked_dialog.hide()

    def refreshTime(self):
        if self.autoDateCheckBox.isChecked():
            dt = QtCore.QDateTime.currentDateTimeUtc()
            self.dateOffEdit.setDate(dt.date())
            self.timeOffEdit.setText(dt.time().toString('HH:mm:ss'))

    def setStartTimeNow(self):
        dt = QtCore.QDateTime.currentDateTimeUtc()
        self.dateOnEdit.setDate(dt.date())
        self.timeOnEdit.setText(dt.time().toString('HH:mm:ss'))
        self.timeOnChanged(self.timeOnEdit.text())

    def autoDateCBChanged(self, checked: bool):
        if not checked:
            self.timeOffEdit.clear()
            self.timeOffChanged('')
        else:
            self.refreshTime()

    def setQSO(self, call: str, band: str = '', freq: float = .0):
        if band:
            self.bandComboBox.setCurrentText(band)
        if freq:
            self.freqDoubleSpinBox.setValue(freq)
        self.callSignLineEdit.setText(call)
        self.callSignChanged(call)

    def clear(self):
        self.__old_values__ = {}
        self.qso_id = None

        if self.__change_mode__:
            self.setChangeMode(False)

        self.callSignLineEdit.clear()
        self.nameLineEdit.clear()
        self.QTHLineEdit.clear()
        self.locatorLineEdit.clear()
        self.RSTSentLineEdit.setText('59')
        self.RSTRcvdLineEdit.setText('59')
        self.commentLineEdit.clear()
        self.remarksTextEdit.clear()

        self.callSignChanged('')
        self.locatorChanged('')
        self.ownCallSignChanged(self.ownCallSignLineEdit.text())
        self.rstSentChanged(self.RSTSentLineEdit.text())
        self.rstRcvdChanged(self.RSTRcvdLineEdit.text())
        self.ownQTHChanged(self.ownQTHComboBox.currentText())

        dt = QtCore.QDateTime.currentDateTimeUtc()
        self.dateOffEdit.setDate(dt.date())
        self.dateOnEdit.setDate(dt.date())
        if self.autoDateCheckBox.isChecked():
            self.timeOffEdit.setText(dt.time().toString('HH:mm:ss'))
        self.timeOnEdit.setText('')
        self.timeOnChanged('')

        if bool(int(self.settings.value('station_cb/cb_by_default', 0))):
            self.bandComboBox.setCurrentText('11m')

        if self.bandComboBox.currentIndex() < 0:
            self.bandComboBox.setCurrentIndex(0)
            self.submodeComboBox.setCurrentIndex(-1)
        if self.modeComboBox.currentIndex() < 0:
            self.modeComboBox.setCurrentIndex(0)

        self.bandComboBox.setPalette(ColorPalettes.PaletteOk)
        self.modeComboBox.setPalette(ColorPalettes.PaletteOk)

        self.qslBurDirGroupBox.setChecked(False)
        self.qslViaLineEdit.clear()
        self.qslBureauRadioButton.setChecked(False)
        self.qslDirectRadioButton.setChecked(False)
        self.qslAccBureauCheckBox.setChecked(False)
        self.qslAccDirectCheckBox.setChecked(False)
        self.qslAcceQSLCheckBox.setChecked(False)
        self.qslAccLoTWCheckBox.setChecked(False)
        self.qslMessageTextEdit.clear()
        self.qslSentCheckBox.setChecked(False)
        self.qslRcvdCheckBox.setChecked(False)

        self.contestComboBox.setCurrentText('')
        self.sentQSOSpinBox.setValue(0)
        self.rcvdQSOSpinBox.setValue(0)
        self.rcvdDataLineEdit.clear()

        self.eventComboBox.setCurrentIndex(0)
        self.txExchLineEdit.clear()
        self.rxExchLineEdit.clear()

        self.toolBox.setCurrentIndex(0)

    def reset(self):
        self.autoDateCheckBox.setEnabled(True)
        self.autoDateCheckBox.setChecked(True)
        self.stationGroupBox.setChecked(True)
        self.identityGroupBox.setChecked(True)

        self.radioComboBox.setCurrentText(self.settings.value('station/radio', ''))
        self.antennaComboBox.setCurrentText(self.settings.value('station/antenna', ''))

    def setChangeMode(self, activate=True):
        self.__change_mode__ = activate
        self.startTimers(not activate)

        if activate:
            self.stationGroupBox.setChecked(False)
            self.stationGroupBox.setCheckable(False)
            self.stationGroupBox.setTitle(self.tr('Station'))
            self.identityGroupBox.setChecked(False)
            self.identityGroupBox.setCheckable(False)
            self.identityGroupBox.setTitle(self.tr('Identity'))
            self.autoDateCheckBox.setChecked(False)
            self.autoDateCheckBox.setEnabled(False)
            self.timeNowPushButton.setEnabled(False)
        else:
            self.stationGroupBox.setCheckable(True)
            self.stationGroupBox.setTitle(self.tr('Configured station'))
            self.stationGroupBox.setChecked(True)
            self.identityGroupBox.setCheckable(True)
            self.identityGroupBox.setTitle(self.tr('Configured identity'))
            self.identityGroupBox.setChecked(True)
            self.autoDateCheckBox.setChecked(True)
            self.autoDateCheckBox.setEnabled(True)
            self.timeNowPushButton.setEnabled(True)

    def bandChanged(self, band: str):
        if band in self.bands:
            self.bandComboBox.setPalette(ColorPalettes.PaletteOk)
            self.freqDoubleSpinBox.setMinimum(self.bands[band][0] - self.bands[band][2])
            self.freqDoubleSpinBox.setValue(self.bands[band][0] - self.bands[band][2])
            self.freqDoubleSpinBox.setMaximum(self.bands[band][1])
            self.freqDoubleSpinBox.setSingleStep(self.bands[band][2])
        elif not band:
            self.bandComboBox.setPalette(ColorPalettes.PaletteEmpty)
        else:
            self.bandComboBox.setPalette(ColorPalettes.PaletteFaulty)
            self.freqDoubleSpinBox.setMinimum(0)
            self.freqDoubleSpinBox.setValue(0)
            self.freqDoubleSpinBox.setMaximum(self.bands['submm'][1])
            self.freqDoubleSpinBox.setSingleStep(1)

        self.modeComboBox.clear()
        if band == '11m':
            self.modeComboBox.insertItems(0, self.modes['CB'].keys())
            self.modeComboBox.setCurrentIndex(0)
            self.powerSpinBox.setMaximum(12)
            self.channelComboBox.setVisible(True)
            self.channelLabel.setVisible(True)
            self.freqDoubleSpinBox.setEnabled(False)
            self.searchHamQTHPushButton.setEnabled(False)
            self.searchQRZCQPushButton.setEnabled(False)
            self.qslPage.setEnabled(False)
            self.contestPage.setEnabled(False)
            self.channelComboBox.setCurrentIndex(-1)
            self.channelComboBox.setCurrentIndex(0)

            if self.stationGroupBox.isChecked():
                self.radioComboBox.setCurrentText(self.settings.value('station_cb/radio', ''))
                self.antennaComboBox.setCurrentText(self.settings.value('station_cb/antenna', ''))

            if self.identityGroupBox.isChecked():
                self.ownCallSignLineEdit.setText(self.settings.value('station_cb/callSign', ''))
        else:
            self.modeComboBox.insertItems(0, self.settings.value('ui/show_modes', self.modes['AFU'].keys()))
            self.modeComboBox.setCurrentIndex(0)
            self.powerSpinBox.setMaximum(1000)
            self.channelComboBox.setVisible(False)
            self.channelLabel.setVisible(False)
            self.freqDoubleSpinBox.setEnabled(True)
            self.searchHamQTHPushButton.setEnabled(True)
            self.searchQRZCQPushButton.setEnabled(True)
            self.qslPage.setEnabled(True)
            self.contestPage.setEnabled(True)

            if self.stationGroupBox.isChecked():
                self.radioComboBox.setCurrentText(self.settings.value('station/radio', ''))
                self.antennaComboBox.setCurrentText(self.settings.value('station/antenna', ''))

            if self.identityGroupBox.isChecked():
                self.ownCallSignLineEdit.setText(self.settings.value('station/callSign', ''))

    def modeChanged(self, mode: str):
        self.submodeComboBox.clear()
        self.submodeComboBox.setEnabled(False)
        if self.bandComboBox.currentText() == '11m':
            if mode in self.modes['CB']:
                self.modeComboBox.setPalette(ColorPalettes.PaletteOk)
                if self.modes['CB'][mode]:
                    self.submodeComboBox.setEnabled(True)
                    self.submodeComboBox.insertItems(0, [''] + self.modes['CB'][mode])
            elif not mode:
                self.modeComboBox.setPalette(ColorPalettes.PaletteEmpty)
            else:
                self.modeComboBox.setPalette(ColorPalettes.PaletteFaulty)
        else:
            if mode in self.modes['AFU']:
                self.modeComboBox.setPalette(ColorPalettes.PaletteOk)
                if self.modes['AFU'][mode]:
                    self.submodeComboBox.setEnabled(True)
                    self.submodeComboBox.insertItems(0, [''] + self.modes['AFU'][mode])
            elif not mode:
                self.modeComboBox.setPalette(ColorPalettes.PaletteEmpty)
            else:
                self.modeComboBox.setPalette(ColorPalettes.PaletteFaulty)

    def stationChanged(self, checked):
        if checked:
            self.ownQTHComboBox.setCurrentText(self.settings.value('station/qth_loc', ''))

            if self.bandComboBox.currentText() == '11m':
                self.radioComboBox.setCurrentText(self.settings.value('station_cb/radio', ''))
                self.antennaComboBox.setCurrentText(self.settings.value('station_cb/antenna', ''))
            else:
                self.radioComboBox.setCurrentText(self.settings.value('station/radio', ''))
                self.antennaComboBox.setCurrentText(self.settings.value('station/antenna', ''))

    def identityChanged(self, checked):
        if checked:
            self.ownNameLineEdit.setText(self.settings.value('station/name', ''))

            if self.bandComboBox.currentText() == '11m':
                self.ownCallSignLineEdit.setText(self.settings.value('station_cb/callSign', ''))
            else:
                self.ownCallSignLineEdit.setText(self.settings.value('station/callSign', ''))

    def channelChanged(self, ch):
        if ch and ch != '-':
            self.freqDoubleSpinBox.setValue(self.cb_channels[ch]['freq'])
        else:
            self.freqDoubleSpinBox.setValue(self.bands['11m'][0] - self.bands['11m'][2])

    def rstRcvdChanged(self, txt):
        if not txt:
            self.RSTRcvdLineEdit.setPalette(ColorPalettes.PaletteEmpty)
        elif check_format(REGEX_RSTFIELD, txt):
            self.RSTRcvdLineEdit.setPalette(ColorPalettes.PaletteOk)
        else:
            self.RSTRcvdLineEdit.setPalette(ColorPalettes.PaletteFaulty)

    def rstSentChanged(self, txt):
        if not txt:
            self.RSTSentLineEdit.setPalette(ColorPalettes.PaletteEmpty)
        elif check_format(REGEX_RSTFIELD, txt):
            self.RSTSentLineEdit.setPalette(ColorPalettes.PaletteOk)
        else:
            self.RSTSentLineEdit.setPalette(ColorPalettes.PaletteFaulty)

    def checkRequired(self):
        if self.callSignLineEdit.text() and self.timeOnEdit.text():
            self.savePushButton.setEnabled(True)
        else:
            self.savePushButton.setEnabled(False)

    def callSignChanged(self, txt):
        self.checkRequired()
        self.setWorkedBefore()
        self.ctyCtyLabel.clear()
        self.ctyAreaLabel.clear()
        if not txt:
            self.callSignLineEdit.setPalette(ColorPalettes.PaletteRequired)
        elif self.bandComboBox.currentText() == '11m':
            self.callSignLineEdit.setPalette(ColorPalettes.PaletteOk)
        elif check_format(REGEX_CALL, txt):
            # worked_dict:dict = self.dragonlog.workedBefore(check_call(txt)[1])
            worked = self.dragonlog.workedBefore(check_call(txt)[1]).keys()
            # for call, data in zip(worked_dict.keys(), worked_dict.values()):
            #     worked.append(f'{call} on {data["date_time"]}')
            if not self.__change_mode__ and worked:
                self.setWorkedBefore(worked)
                self.callSignLineEdit.setPalette(ColorPalettes.PaletteWorked)
            else:
                self.callSignLineEdit.setPalette(ColorPalettes.PaletteOk)

            cdata: Country = self.dragonlog.cty_data(txt)
            if cdata:
                self.ctyCtyLabel.setText(f'{cdata.code} {cdata.name}, {cdata.continent}')
                self.ctyAreaLabel.setText(f'DXCC={cdata.dxcc}, CQ={cdata.cq}, ITU={cdata.itu}')
        else:
            self.callSignLineEdit.setPalette(ColorPalettes.PaletteFaulty)

    def locatorChanged(self, txt):
        if not txt:
            self.locatorLineEdit.setPalette(ColorPalettes.PaletteEmpty)
            self.distLabel.clear()
        elif check_format(REGEX_LOCATOR, txt):
            self.locatorLineEdit.setPalette(ColorPalettes.PaletteOk)
            # noinspection PyBroadException
            try:
                _, own_locator = check_qth(self.ownQTHComboBox.currentText())
                self.distLabel.setText(f'{distance(txt, own_locator)} km')
            except Exception:
                self.distLabel.clear()
        else:
            self.locatorLineEdit.setPalette(ColorPalettes.PaletteFaulty)
            self.distLabel.clear()

    def ownCallSignChanged(self, txt):
        if not txt:
            self.ownCallSignLineEdit.setPalette(ColorPalettes.PaletteEmpty)
        elif check_format(REGEX_CALL, txt):
            self.ownCallSignLineEdit.setPalette(ColorPalettes.PaletteOk)
        elif self.bandComboBox.currentText() == '11m':
            self.ownCallSignLineEdit.setPalette(ColorPalettes.PaletteOk)
        else:
            self.ownCallSignLineEdit.setPalette(ColorPalettes.PaletteFaulty)

    def ownQTHChanged(self, txt):
        if not txt:
            self.ownQTHComboBox.setPalette(ColorPalettes.PaletteEmpty)
        elif check_qth(txt.strip()):
            self.ownQTHComboBox.setPalette(ColorPalettes.PaletteOk)
        else:
            self.ownQTHComboBox.setPalette(ColorPalettes.PaletteFaulty)

    def timeOnChanged(self, txt):
        self.checkRequired()
        if not txt:
            self.timeOnEdit.setPalette(ColorPalettes.PaletteRequired)
        elif check_format(REGEX_TIME, txt):
            self.timeOnEdit.setPalette(ColorPalettes.PaletteOk)
        else:
            self.timeOnEdit.setPalette(ColorPalettes.PaletteFaulty)

    def timeOffChanged(self, txt):
        if not txt:
            self.timeOffEdit.setPalette(ColorPalettes.PaletteEmpty)
        elif check_format(REGEX_TIME, txt):
            self.timeOffEdit.setPalette(ColorPalettes.PaletteOk)
        else:
            self.timeOffEdit.setPalette(ColorPalettes.PaletteFaulty)

    @property
    def values(self) -> tuple:
        """Retreiving all values from the form"""

        date_time_off = ''
        if check_format(REGEX_TIME, self.timeOffEdit.text()):
            date_time_off = self.dateOffEdit.text() + ' ' + self.timeOffEdit.text()
        else:
            if self.timeOffEdit.text():
                self.log.warning(f'Wrong time format for end time')

        if check_format(REGEX_TIME, self.timeOnEdit.text()):
            date_time_on = self.dateOnEdit.text() + ' ' + self.timeOnEdit.text()
        else:
            self.log.warning(f'Wrong time format for start time. Fallback to current date time.')
            date_time_on = QtCore.QDateTime.currentDateTimeUtc().toString('yyyy-MM-dd HH:mm:ss')

        band = self.bandComboBox.currentText()

        prop = ''
        if self.propComboBox.currentText():
            prop = self.prop_trans[self.propComboBox.currentText()]

        own_qth = self.ownQTHComboBox.currentText()
        own_locator = ''
        if check_qth(self.ownQTHComboBox.currentText()):
            own_qth, own_locator = check_qth(self.ownQTHComboBox.currentText())

        qsl_path = ''
        qsl_sent = 'N'
        qsl_rcvd = 'N'
        if self.qslBurDirGroupBox.isChecked():
            qsl_path = 'D' if self.qslDirectRadioButton.isChecked() else 'B'

            if self.qslBurDirGroupBox.isChecked():
                qsl_sent = 'Y' if self.qslSentCheckBox.isChecked() else 'N'
                qsl_rcvd = 'Y' if self.qslRcvdCheckBox.isChecked() else 'N'

        # noinspection PyBroadException
        try:
            dist = distance(self.locatorLineEdit.text(), own_locator)
        except Exception:
            dist = 0

        return (
            date_time_on,
            date_time_off,
            self.ownCallSignLineEdit.text().strip().upper() if band != '11m' else self.ownCallSignLineEdit.text().strip(),
            self.callSignLineEdit.text().strip().upper() if band != '11m' else self.callSignLineEdit.text().strip(),
            self.nameLineEdit.text().strip(),
            self.QTHLineEdit.text().strip(),
            self.locatorLineEdit.text().strip(),
            self.RSTSentLineEdit.text().strip(),
            self.RSTRcvdLineEdit.text().strip(),
            band,
            self.modeComboBox.currentText().strip(),
            self.submodeComboBox.currentText().strip(),
            self.freqDoubleSpinBox.value() if self.freqDoubleSpinBox.value() > self.freqDoubleSpinBox.minimum() else '',
            self.channelComboBox.currentText() if band == '11m' else '-',
            self.powerSpinBox.value() if self.powerSpinBox.value() > 0 else '',
            prop,
            self.ownNameLineEdit.text().strip(),
            own_qth,
            own_locator,
            self.radioComboBox.currentText().strip(),
            self.antennaComboBox.currentText().strip(),
            self.remarksTextEdit.toPlainText().strip(),
            self.commentLineEdit.text().strip(),
            dist,
            self.qslViaLineEdit.text(),
            qsl_path,
            self.qslMessageTextEdit.toPlainText().strip(),
            qsl_sent,
            qsl_rcvd,
            self.__old_values__.get('eqsl_sent', 'N'),
            self.__old_values__.get('eqsl_rcvd', 'N'),
            self.__old_values__.get('lotw_sent', 'N'),
            self.__old_values__.get('lotw_rcvd', 'N'),
            self.__old_values__.get('hamqth', 'N'),
            CONTEST_IDS.get(self.contestComboBox.currentText().strip(), self.contestComboBox.currentText().strip()),
            self.sentQSOSpinBox.value() if self.contestComboBox.currentText().strip() else 0,
            self.rcvdQSOSpinBox.value() if self.contestComboBox.currentText().strip() else 0,
            self.rcvdDataLineEdit.text().strip() if self.contestComboBox.currentText().strip() else '',
            self.eventComboBox.currentText(),
            self.txExchLineEdit.text().strip() if self.eventComboBox.currentText().strip() else '',
            self.rxExchLineEdit.text().strip() if self.eventComboBox.currentText().strip() else '',
        )

    @values.setter
    def values(self, values: dict):
        """Set all form values"""
        self.__old_values__ = values.copy()

        self.qso_id = values['id']
        time_on = ''
        if values['date_time']:
            try:
                date_on, time_on = values['date_time'].split()
                self.dateOnEdit.setDate(QtCore.QDate.fromString(date_on, 'yyyy-MM-dd'))
            except ValueError:
                self.log.error(f'Wrong date time format for start time in QSO #{self.qso_id}')
        self.timeOnEdit.setText(time_on)
        self.timeOnChanged(self.timeOnEdit.text())

        time_off = ''
        if values['date_time_off']:
            try:
                date_off, time_off = values['date_time_off'].split()
                self.dateOffEdit.setDate(QtCore.QDate.fromString(date_off, 'yyyy-MM-dd'))
            except ValueError:
                self.log.error(f'Wrong date time format for end time in QSO #{self.qso_id}')
        self.timeOffEdit.setText(time_off)
        self.timeOffChanged(self.timeOffEdit.text())

        band = values['band']
        self.bandComboBox.setCurrentText(band)

        self.ownCallSignLineEdit.setText(values['own_callsign'])
        self.callSignLineEdit.setText(values['call_sign'])
        self.nameLineEdit.setText(values['name'])
        self.QTHLineEdit.setText(values['qth'])
        self.locatorLineEdit.setText(values['locator'])
        self.RSTSentLineEdit.setText(values['rst_sent'])
        self.RSTRcvdLineEdit.setText(values['rst_rcvd'])

        self.callSignChanged(self.callSignLineEdit.text())
        self.locatorChanged(self.locatorLineEdit.text())
        self.ownCallSignChanged(self.ownCallSignLineEdit.text())
        self.rstSentChanged(self.RSTSentLineEdit.text())
        self.rstRcvdChanged(self.RSTRcvdLineEdit.text())
        self.ownQTHChanged(self.ownQTHComboBox.currentText())

        self.modeComboBox.setCurrentText(values['mode'])
        if values['submode']:
            self.submodeComboBox.setCurrentText(values['submode'])

        freq = 0
        try:
            freq = float(values['freq'])
        except ValueError:
            if band in self.bands:
                freq = self.bands[band][0] - self.bands[band][2]
        self.freqDoubleSpinBox.setValue(freq)

        if band == '11m':
            self.channelComboBox.setCurrentIndex(-1)
            channel = values['channel']
            self.channelComboBox.setCurrentText(str(channel) if channel else '-')

        try:
            power = int(values['power'])
        except ValueError:
            power = 0
        self.powerSpinBox.setValue(power)

        if values['propagation']:
            self.propComboBox.setCurrentText(self.prop[values['propagation']])
        else:
            self.propComboBox.setCurrentIndex(0)

        self.ownNameLineEdit.setText(values['own_name'])

        if values['own_qth'] and values['own_locator']:
            self.ownQTHComboBox.setCurrentText(f"{values['own_qth']} ({values['own_locator']})")
        elif values['own_qth']:
            self.ownQTHComboBox.setCurrentText(values['own_qth'])
        elif values['own_locator']:
            self.ownQTHComboBox.setCurrentText(f"({values['own_locator']})")

        self.radioComboBox.setCurrentText(values['radio'])
        self.antennaComboBox.setCurrentText(values['antenna'])
        self.remarksTextEdit.setText(values['remarks'])
        self.commentLineEdit.setText(values['comments'].replace('\n', ' ').replace('\r', ''))

        self.qslViaLineEdit.setText(values['qsl_via'])
        self.qslMessageTextEdit.setText(values['qsl_msg'])
        if values['qsl_sent'] in ('R', 'Y') or values['qsl_rcvd'] in ('R', 'Y'):
            self.qslBureauRadioButton.setChecked(values['qsl_path'] == 'B')
            self.qslDirectRadioButton.setChecked(values['qsl_path'] == 'D')

            if values['qsl_sent'] in ('R', 'Y') or values['qsl_rcvd'] in ('R', 'Y'):
                self.qslBurDirGroupBox.setChecked(True)
                self.qslSentCheckBox.setChecked(values['qsl_sent'] == 'Y')
                self.qslRcvdCheckBox.setChecked(values['qsl_rcvd'] == 'Y')

        self.contestComboBox.setCurrentText(CONTEST_NAMES.get(values['contest_id'], ''))
        self.rcvdDataLineEdit.setText(values['crx_data'])
        try:
            self.sentQSOSpinBox.setValue(int(values['ctx_qso_id']) if values['ctx_qso_id'] else -1)
        except ValueError:
            self.log.warning(f'Sent QSO ID "{values["ctx_qso_id"]}" is not a number')

        try:
            self.rcvdQSOSpinBox.setValue(int(values['crx_qso_id']) if values['crx_qso_id'] else -1)
        except ValueError:
            self.log.warning(f'Rcvd QSO ID "{values["crx_qso_id"]}" is not a number')
            if not values['crx_data']:
                self.rcvdDataLineEdit.setText(values['crx_qso_id'])
                self.log.info(f'Stored QSO ID as rcvd data instead')

        self.eventComboBox.setCurrentText(values['event'])
        self.txExchLineEdit.setText(values['evt_tx_exch'])
        self.rxExchLineEdit.setText(values['evt_rx_exch'])

    def searchHamQTH(self):
        self.searchCallbook(self.callbook_hamqth)

    def searchQRZCQ(self):
        self.searchCallbook(self.callbook_qrzcq)

    def searchQRZ(self):
        self.searchCallbook(self.callbook_qrz)

    def searchCallbook(self, callbook):
        if not self.callSignLineEdit.text().strip():
            return

        try:
            if not callbook.is_loggedin:
                callbook.login(self.settings.value(f'callbook/{callbook.callbook_type.name}_user', ''),
                               self.settings_form.callbookPassword(callbook.callbook_type))
                self.log.info(f'Logged into callbook {callbook.callbook_type.name}')

            data: CallBookData | None = None
            for _ in range(2):
                try:
                    data = callbook.get_dataset(self.callSignLineEdit.text())
                    break
                except SessionExpiredException:
                    self.log.debug('Callbook session expired')
                    callbook.login(self.settings.value(f'callbook/{callbook.callbook_type.name}_user', ''),
                                   self.settings_form.callbookPassword(callbook.callbook_type))

            if data:
                if data.nickname and not self.nameLineEdit.text().strip():
                    self.nameLineEdit.setText(data.nickname)
                if data.locator and not self.locatorLineEdit.text().strip():
                    self.locatorLineEdit.setText(data.locator)
                    self.locatorChanged(data.locator)
                if data.qth and not self.QTHLineEdit.text().strip():
                    self.QTHLineEdit.setText(data.qth)
                if data.qsl_via and not self.qslViaLineEdit.text().strip():
                    self.qslViaLineEdit.setText(data.qsl_via)

                self.qslAccBureauCheckBox.setChecked(data.qsl_bureau)
                self.qslAccDirectCheckBox.setChecked(data.qsl_direct)
                self.qslAcceQSLCheckBox.setChecked(data.qsl_eqsl)
                self.qslAccLoTWCheckBox.setChecked(data.qsl_lotw)

                self.log.info(f'Fetched data from callbook {callbook.callbook_type.name}')
        except LoginException:
            QtWidgets.QMessageBox.warning(self, self.tr('Callbook search error'),
                                          self.tr('Login failed for user') + ': ' + self.settings.value(
                                              f'callbook/{callbook.callbook_type.name}_user', ''))
        except CallsignNotFoundException as exc:
            QtWidgets.QMessageBox.information(self, self.tr('Callbook search result'),
                                              self.tr('Callsign not found') + f': {exc.args[0]}')
        except Exception as exc:
            self.log.exception(exc)
            QtWidgets.QMessageBox.warning(self, self.tr('Callbook search error'),
                                          self.tr('During callbook search an error occured') + f':\n{exc}')

    def saveLog(self):
        if self.__change_mode__:
            self.dragonlog.updateQSO(self.qso_id)
        else:
            self.dragonlog.fetchQSO()

        self.toolBox.setCurrentIndex(0)
        self.clear()

    def contestChanged(self, text: str):
        self.xotaGroupBox.setDisabled(bool(text))
        contest: ContestLog | None = CONTESTS.get(CONTEST_IDS.get(text, ''), None)
        if contest:
            self.rcvdDataLineEdit.setPlaceholderText(contest.contest_exch_fmt)
        else:
            self.rcvdDataLineEdit.setPlaceholderText(self.tr('Rx Exchange'))

    def eventChanged(self, text: str):
        self.contestGroupBox.setDisabled(bool(text))

    def keyPressEvent(self, e):
        if e.key() != QtCore.Qt.Key.Key_Escape:
            super().keyPressEvent(e)



# DragonLog (c) 2023-2025 by Andreas Schawo is licensed under CC BY-SA 4.0.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/
"""Base definitions for contest processing"""

import os
import re
import typing
from enum import Enum, auto
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass

from dragonlog.RegEx import check_format, REGEX_CALL, REGEX_LOCATOR, REGEX_RSTFIELD
from dragonlog.distance import distance

# noinspection SpellCheckingInspection
MODE_MAP_CBR = {
    'CW': 'CW',
    'SSB': 'PH',
    'FM': 'FM',
    'RTTY': 'RY',
    'FT8': 'DG',
    'MFSK': 'DG',
}

MODE_MAP_EDI = {
    'UNKNOWN': 0,
    'SSB': 1,
    'PH': 1,
    'CW': 2,
    'SSB-CW': 3,
    'CB-SSB': 4,
    'AM': 5,
    'FM': 6,
    'RTTY ': 7,
    'RY': 7,
    'SSTV': 8,
    'ATV': 9,
}

BAND_MAP_CBR = {
    'all': 'ALL',
    '160m': '1800',
    '80m': '3500',
    '40m': '7000',
    '20m': '14000',
    '15m': '21000',
    '10m': '28000',
    '6m': '50',
    '4m': '70',
    '2m': '144',
    '1.25m': '222',
    '70cm': '432',
    '33cm': '902',
    '23cm': '1.2G',
    '13cm': '2.3G',
    '9cm': '3.4G',
    '6cm': '5.7G',
    '3cm': '10G',
    '1.25cm': '24G',
    '6mm': '47G',
    '4mm': '75G',
    '2.5mm': '122G',
    '2mm': '134G',
    '1mm': '241G',
    # 'submm': 'LIGHT',
}

BAND_MAP_EDI = {
    '6m': '50 MHz',
    '4m': '70 MHz',
    '2m': '144 MHz',
    '70cm': '432 MHz',
    '23cm': '1.3 GHz',
    '13cm': '2.3 GHz',
    '9cm': '3.4 GHz',
    '6cm': '5.7 GHz',
    '3cm': '10 GHz',
    '1.25cm': '24 GHz',
    '6mm': '47 GHz',
    '4mm': '75 GHz',
    '2.5mm': '120 GHz',
    '2mm': '144 GHz',
    '1mm': '248 GHz',
}

BAND_FROM_CBR = dict(zip(BAND_MAP_CBR.values(), BAND_MAP_CBR.keys()))


@dataclass
class Record:
    """Represents data for a QSO record"""
    band: str
    mode: str
    date: str
    time: str
    own_call: str
    sent_rst: str
    sent_exch: str
    call: str
    rcvd_rst: str
    rcvd_exch: str
    tx: str


@dataclass
class CBRRecord(Record):
    """Represents data for a QSO record in CBR format"""
    pass


@dataclass
class EDIRecord(CBRRecord):
    """Represents data for a QSO record in EDI format"""
    dist: int


class CategoryBand(Enum):
    B_ALL = auto()
    B_160M = auto()
    B_80M = auto()
    B_40M = auto()
    B_20M = auto()
    B_15M = auto()
    B_10M = auto()
    B_6M = auto()
    B_4M = auto()
    B_2M = auto()
    B_1_25M = auto()
    B_70CM = auto()
    B_23CM = auto()
    B_13CM = auto()
    B_9CM = auto()
    B_6CM = auto()
    B_3CM = auto()
    B_1_25CM = auto()
    B_6MM = auto()
    B_4MM = auto()
    B_2_5MM = auto()
    B_2MM = auto()
    B_1MM = auto()
    B_SUBMM = auto()

    def __str__(self):
        return self.name[2:].lower().replace('_', '.')

    @classmethod
    def from_str(cls, name: str):
        return cls['B_' + str(name).upper().replace('.', '_')]


class CategoryEnum(Enum):
    def __str__(self):
        return self.name.replace('_', '-')

    @classmethod
    def from_str(cls, name: str):
        return cls[str(name).upper().replace('-', '_')]


class CategoryOperator(CategoryEnum):
    SINGLE_OP = auto()
    MULTI_OP = auto()
    CHECKLOG = auto()
    TRAINEE = auto()


class CategoryMode(CategoryEnum):
    CW = auto()
    SSB = auto()
    RTTY = auto()
    MIXED = auto()
    DIGI = auto()
    FM = auto()


class CategoryPower(CategoryEnum):
    LOW = auto()
    HIGH = auto()
    QRP = auto()
    A = auto()
    B = auto()
    C = auto()
    NONE = auto()


class CategoryAssisted(CategoryEnum):
    ASSISTED = auto()
    NON_ASSISTED = auto()


class CategoryTransmitter(CategoryEnum):
    ONE = auto()
    TWO = auto()
    LIMITED = auto()
    UNLIMITED = auto()
    SWL = auto()


@dataclass
class Address:
    street: str
    zip: str
    city: str
    country: str


@dataclass
class BandStatistics:
    qsos: int = 0
    rated: int = 0
    points: int = 0
    multis: int = 1
    multis2: int = 0
    summary: int = 0

    def values(self) -> tuple:
        return tuple(self.__dict__.values())


@dataclass
class ExchangeData:
    """Represents exchange data in a reliable structure"""
    number: str = ''
    locator: str = ''
    power: str = ''
    darc_dok: str = ''
    itu_zone: str = ''


class InvalidBandException(Exception):
    pass


class InvalidModeException(Exception):
    pass


class ContestLog(ABC):
    """Abstract base class for ham radio contests"""
    contest_name = 'Contest'
    contest_year = '2025'
    contest_update = '2025-01-01'
    contest_exch_fmt = 'None'

    REGEX_TIME = re.compile(r'(([0-1][0-9])|(2[0-3]))([0-5][0-9])([0-5][0-9])?')
    REGEX_DATE = re.compile(r'([1-9][0-9]{3})((0[1-9])|(1[0-2]))((0[1-9])|([1-2][0-9])|(3[0-2]))')

    def __init__(self, callsign: str, name: str, club: str, address: Address, email: str, locator: str,
                 band: type[CategoryBand], mode: type[CategoryMode],
                 pwr: type[CategoryPower] = CategoryPower.HIGH,
                 cat_operator: type[CategoryOperator] = CategoryOperator.SINGLE_OP,
                 assisted: type[CategoryAssisted] = CategoryAssisted.NON_ASSISTED,
                 tx: type[CategoryTransmitter] = CategoryTransmitter.ONE,
                 operators: list[str] = None, specific: str = '', skip_id: bool = False,
                 skip_warn: bool = False, logger=None, **params):

        self.log = logging.getLogger(type(self).__name__)
        if logger:
            self.log.setLevel(logger.loglevel)
            self.log.addHandler(logger)
        self.log.debug('Initialising...')

        self.__header__: dict[str, str] = {
            'CONTEST': '',  # Class
            'CALLSIGN': '',
            'CATEGORY-OPERATOR': '',
            'CATEGORY-BAND': '',
            'CATEGORY-MODE': '',
            'CATEGORY-POWER': '',
            'CATEGORY-ASSISTED': '',
            'CATEGORY-TRANSMITTER': '',
            'SPECIFIC': '',
            'CLAIMED-SCORE': '',  # Class
            'CREATED-BY': '',  # Modul
            'OPERATORS': '',
            'NAME': '',
            'CLUB': '',
            'ADDRESS': '',
            'EMAIL': '',
            'GRID-LOCATOR': '',
        }
        self.__init_header__(callsign, name, club, address, email, locator,
                             band, mode, pwr, cat_operator, assisted, tx, operators, specific)

        self.__qso_id__: int = 0

        self.__errors__: int = 0
        self.__warnings__: int = 0
        self.__infos__: int = 0

        self.__qsos__: list[Record] = []

        self.__rated__: int = 0
        self.__points__: int = 0
        self.__multis__: set[str] = set()
        self.__multis2__: set[str] = set()

        valid_bands = self.valid_bands_list().copy()
        if 'all' in valid_bands:
            valid_bands.remove('all')
        self.__stats__: dict[str, BandStatistics] = dict(zip(valid_bands,
                                                             [self.statistic_primer() for _ in
                                                              range(len(valid_bands))]))
        self.__skip_id__ = skip_id
        self.__skip_warn__ = skip_warn

        self.__contest_date__ = ''

        self.__out_file__ = None

    def __init_header__(self, callsign, name, club, address: Address, email, locator,
                        band: type[CategoryBand], mode: type[CategoryMode],
                        pwr: type[CategoryPower] = CategoryPower.HIGH,
                        cat_operator: type[CategoryOperator] = CategoryOperator.SINGLE_OP,
                        assisted: type[CategoryAssisted] = CategoryAssisted.NON_ASSISTED,
                        tx: type[CategoryTransmitter] = CategoryTransmitter.ONE,
                        operators: list[str] = None, specific: str = ''):

        if not check_format(REGEX_CALL, callsign):
            self.log.error(f'Callsign "{callsign}" does not match call format')
        if locator and not check_format(REGEX_LOCATOR, locator):
            self.log.warning(f'Locator "{locator}" does not match locator format')

        self.__header__['CONTEST'] = self.contest_id
        self.__header__['CREATED-BY'] = 'ContestLog v0.1'
        self.__header__['CALLSIGN'] = callsign
        self.__header__['NAME'] = name
        self.__header__['CLUB'] = club
        self.__header__['ADDRESS'] = address.street
        self.__header__['ADDRESS-POSTALCODE'] = address.zip
        self.__header__['ADDRESS-CITY'] = address.city
        self.__header__['ADDRESS-COUNTRY'] = address.country
        self.__header__['EMAIL'] = email
        self.__header__['GRID-LOCATOR'] = locator
        self.__header__['OPERATORS'] = ' '.join(operators) if operators else callsign
        self.__header__['CATEGORY-OPERATOR'] = str(cat_operator)
        self.__header__['CATEGORY-BAND'] = str(band)
        self.__header__['CATEGORY-MODE'] = str(mode)
        self.__header__['CATEGORY-POWER'] = str(pwr)
        self.__header__['CATEGORY-ASSISTED'] = str(assisted)
        self.__header__['CATEGORY-TRANSMITTER'] = str(tx)
        self.__header__['SPECIFIC'] = specific

    @property
    def contest_id(self) -> str:
        return self.contest_name

    @property
    def infos(self):
        """Number of informations"""
        return self.__infos__

    @property
    def warnings(self):
        """Number of warnings"""
        return self.__warnings__

    @property
    def errors(self):
        """Number of errors and exceptions"""
        return self.__errors__

    def info(self, txt):
        """Logs an information message"""
        self.__infos__ += 1
        self.log.info(f'QSO #{self.__qso_id__}: {txt}')

    def warning(self, txt):
        """Logs a warning message"""
        self.__warnings__ += 1
        self.log.warning(f'QSO #{self.__qso_id__}: {txt}')

    def error(self, txt):
        """Logs an error message"""
        self.__errors__ += 1
        self.log.error(f'QSO #{self.__qso_id__}: {txt}')

    def exception(self, txt='Exception'):
        """Logs an exception message"""
        self.__errors__ += 1
        self.log.exception(f'QSO #{self.__qso_id__}: {txt}')

    def set_created_by(self, program_name: str):
        """Set the created by header value for the program in use"""
        self.__header__['CREATED-BY'] = program_name

    def check_band(self, adif_rec: dict[str, str]) -> bool:
        """Test if the ADIF record uses an valid band"""
        if (adif_rec['BAND'].lower() != self.__header__['CATEGORY-BAND'] and
                BAND_MAP_CBR[adif_rec['BAND'].lower()] != self.__header__['CATEGORY-BAND'].lower()):
            self.log.warning(f'QSO #{self.__qso_id__} band "{adif_rec["BAND"].lower()}" does not match with '
                             f'contest band "{self.__header__["CATEGORY-BAND"]}"')
            if self.__skip_warn__:
                return False
        return True

    def append(self, adif_rec: dict[str, str]):
        """Append an ADIF record to the contest"""
        self.__qso_id__ = adif_rec['APP_DRAGONLOG_QSOID']
        try:
            if self.__skip_id__:
                if 'CONTEST_ID' not in adif_rec:
                    self.error('Contest ID missing. Skipping.')
                    return
                if adif_rec['CONTEST_ID'] != self.__header__['CONTEST']:
                    self.warning(f'Contest ID "{adif_rec["CONTEST_ID"]}" does not match '
                                 f'"{self.__header__["CONTEST"]}". Skipping.')
                    return

            if self.__header__['CATEGORY-BAND'] != 'all' and not self.check_band(adif_rec):
                return

            if self.__header__['CATEGORY-MODE'] != 'MIXED' and adif_rec['MODE'].upper() != self.__header__[
                'CATEGORY-MODE']:
                self.warning(f'Mode "{adif_rec["MODE"].upper()}" does not match with '
                             f'contest mode "{self.__header__["CATEGORY-MODE"]}"')
                if self.__skip_warn__:
                    return

            if not check_format(REGEX_CALL, adif_rec['CALL']):
                self.warning(f'Call "{adif_rec["CALL"]}" does not match call format')
                if self.__skip_warn__:
                    return

            if not check_format(REGEX_RSTFIELD, adif_rec['RST_RCVD']):
                self.warning(
                    f'Rreceived RST "{adif_rec["RST_RCVD"]}" does not match RST format')
                if self.__skip_warn__:
                    return

            if not check_format(REGEX_RSTFIELD, adif_rec['RST_SENT']):
                self.warning(f'Sent RST "{adif_rec["RST_SENT"]}" does not match RST format')
                if self.__skip_warn__:
                    return

            if not check_format(self.REGEX_DATE, adif_rec['QSO_DATE']):
                self.warning(f'Date "{adif_rec["QSO_DATE"]}" does not match date format')
                if self.__skip_warn__:
                    return

            if not check_format(self.REGEX_TIME, adif_rec['TIME_ON']):
                self.warning(f'Time "{adif_rec["TIME_ON"]}" does not match time format')
                if self.__skip_warn__:
                    return

            rec = self.build_record(adif_rec)

            if rec:
                self.process_points(rec)
                self.__qsos__.append(rec)

                self.__header__['CLAIMED-SCORE'] = str(self.claimed_points)

                self.log.debug(self.summary())
        except KeyError as exc:
            self.error(f'Could not be processed. Field {exc} is missing')
        except InvalidBandException as exc:
            self.error(f'Could not be processed. Invalid band {exc}')
        except InvalidModeException as exc:
            self.error(f'Could not be processed. Invalid mode {exc}')

    @abstractmethod
    def build_record(self, adif_rec) -> Record:
        """Build the QSO info"""
        pass

    def add_soapbox(self, txt: str):
        """Add soapbox content or remarks"""
        self.__header__['SOAPBOX'] = str(txt)

    def open_file(self, path: str = os.path.curdir):
        """Open the contest export file"""
        self.__out_file__ = open(os.path.join(path, self.file_name), 'w')

    @abstractmethod
    def write_records(self):
        """Writes the exported content to the open file"""
        pass

    def close_file(self):
        """Close the contest export file"""
        if self.__out_file__:
            self.__out_file__.close()

    @property
    def points(self) -> int:
        """Basic QSO points"""
        return self.__points__

    @property
    def claimed_points(self) -> int:
        """Points with some kind of multiplicator"""
        return self.points * (self.multis + self.multis2)

    @property
    def qsos(self) -> int:
        """Number of QSOs"""
        return len(self.__qsos__)

    @property
    def rated(self) -> int:
        """Number of QSOs rated with points"""
        return self.__rated__

    @property
    def multis(self) -> int:
        """Number of gained multis"""
        return len(self.__multis__) if self.__multis__ else 1

    @property
    def multis2(self) -> int:
        """Number of gained multis2"""
        return len(self.__multis2__)

    @property
    def multis_set(self) -> set[str]:
        """The multis as a set"""
        return self.__multis__

    @property
    def multis2_set(self) -> set[str]:
        """The multis2 as a set"""
        return self.__multis2__

    @abstractmethod
    def process_points(self, rec: Record):
        """Place for calculating points and decision for rating a QSO"""
        pass

    def summary(self) -> str:
        """A summary of points for the contest"""
        return (f'QSOs: {self.qsos}, Rated: {self.rated}, Points: {self.points}, '
                f'Multis: {self.multis}, Multis2: {self.multis2}, Claimed points: {self.claimed_points}')

    @property
    def statistics(self) -> dict[str, BandStatistics]:
        """A detailed statistic for the contest"""
        stats = self.__stats__.copy()
        for b in stats:
            stats[b].multis = self.multis
            stats[b].multis2 = self.multis2
            stats[b].summary = stats[b].points * (self.multis + self.multis2)
        stats['Total'] = BandStatistics(self.qsos, self.rated, self.points,
                                        self.multis, self.multis2, self.claimed_points)
        return stats

    @classmethod
    def statistic_primer(cls) -> BandStatistics:
        return BandStatistics()

    @property
    def file_name(self) -> str:
        """The contest export file name"""
        return f'{self.__contest_date__}_{self.__header__["CALLSIGN"]}-{self.__header__["CATEGORY-BAND"]}.cbr'

    @classmethod
    def valid_bands(cls) -> tuple[CategoryBand, ...]:
        """Valid bands for the contest"""
        return CategoryBand.B_ALL,

    @classmethod
    def valid_bands_list(cls) -> list[str]:
        """Valid bands for the contest as list of strings"""
        return [str(b) for b in cls.valid_bands()]

    @classmethod
    def valid_modes(cls) -> tuple[CategoryMode, ...]:
        """Valid modes for the contest"""
        return CategoryMode.MIXED,

    @classmethod
    def valid_modes_list(cls) -> list[str]:
        """Valid modes for the contest as list of strings"""
        return [m.name for m in cls.valid_modes()]

    @classmethod
    def valid_power(cls) -> tuple[CategoryPower, ...]:
        """Valid power classes for the contest"""
        return CategoryPower.HIGH, CategoryPower.LOW, CategoryPower.QRP

    @classmethod
    def valid_power_list(cls) -> list[str]:
        """Valid power classes for the contest as list of strings"""
        return [p.name for p in cls.valid_power()]

    @classmethod
    def valid_operator(cls) -> tuple[CategoryOperator, ...]:
        """Valid operator classes for the contest"""
        return CategoryOperator.SINGLE_OP, CategoryOperator.MULTI_OP, CategoryOperator.CHECKLOG

    @classmethod
    def valid_operator_list(cls) -> list[str]:
        """Valid operator classes for the contest as list of strings"""
        return [str(o) for o in cls.valid_operator()]

    @classmethod
    def descr_specific(cls) -> str:
        """A description of the contest specific field"""
        return 'Specific'

    @classmethod
    def needs_specific(cls) -> bool:
        """If the specific field is needed to calculate points"""
        return False

    @classmethod
    def is_single_day(cls) -> bool:
        """If the contest is on a single day"""
        return True

    @property
    def exchange_format(self) -> str:
        """Get a textual descritpion of the exchange format expected by the contest"""
        return self.contest_exch_fmt

    @staticmethod
    @abstractmethod
    def extract_exchange(exchange: str) -> ExchangeData | None:
        """Provides structured exchange data to from an exchange string
        Useful to extract locator and other info for logbook"""
        pass

    @staticmethod
    @abstractmethod
    def prepare_exchange(exchange: ExchangeData) -> str:
        """Builds exchange from structured exchange data
        Useful for preparing the exchange from historical database info"""
        pass


class ContestLogCBR(ContestLog):
    """Abstract base class for ham radio contests with an CBR result file (usually shortwave)"""
    contest_name = 'KW-Contest'
    contest_exch_fmt = 'Number'

    def __init__(self, callsign: str, name: str, club: str, address: Address, email: str, locator: str,
                 band: type[CategoryBand], mode: type[CategoryMode],
                 pwr: type[CategoryPower] = CategoryPower.HIGH,
                 cat_operator: type[CategoryOperator] = CategoryOperator.SINGLE_OP,
                 assisted: type[CategoryAssisted] = CategoryAssisted.NON_ASSISTED,
                 tx: type[CategoryTransmitter] = CategoryTransmitter.ONE,
                 operators: list[str] = None, specific: str = '', skip_id: bool = False,
                 skip_warn: bool = False, logger=None, **params):
        super().__init__(callsign, name, club, address, email, locator,
                         band, mode, pwr, cat_operator,
                         assisted, tx, operators, specific, skip_id, skip_warn, logger)

        self.__calls__ = []

    def build_record(self, adif_rec) -> CBRRecord:
        """Build the QSO info"""
        self.log.debug(adif_rec)

        date = f'{adif_rec["QSO_DATE"][:4]}-{adif_rec["QSO_DATE"][4:6]}-{adif_rec["QSO_DATE"][6:8]}'
        if not self.__contest_date__:
            self.__contest_date__ = date

        try:
            band = BAND_MAP_CBR[adif_rec['BAND'].lower()]
        except KeyError as exc:
            raise InvalidBandException(str(exc)) from None

        try:
            mode = MODE_MAP_CBR[adif_rec['MODE']]
        except KeyError as exc:
            raise InvalidModeException(str(exc)) from None

        return CBRRecord(band,
                         mode,
                         date,
                         adif_rec['TIME_ON'][:4],
                         adif_rec['STATION_CALLSIGN'],
                         adif_rec['RST_SENT'],
                         adif_rec.get('STX_STRING', str(adif_rec['STX'])).upper(),
                         adif_rec['CALL'],
                         adif_rec['RST_RCVD'],
                         adif_rec.get('SRX_STRING',
                                      str(adif_rec.get('SRX', 0))).upper().strip().replace(',', ' ').replace('_', ' '),
                         '0'
                         )

    def write_records(self):
        if self.__out_file__:
            for row in self.serialize_cbr():
                self.__out_file__.write(row)
                self.__out_file__.write('\n')

    def serialize_cbr(self) -> typing.Iterator[str]:
        """Serialize the whole CBR content"""

        # CBR format from http://wwrof.org/cabrillo/
        yield 'START-OF-LOG: 3.0'

        for k in self.__header__:
            if k in ('ADDRESS', 'SOAPBOX') and self.__header__[k]:
                for l in self.__header__[k].split('\n'):
                    yield f'{k}: {l}'
            elif self.__header__[k]:
                yield f'{k}: {self.__header__[k]}'

        yield ''  # Divide between header and records

        yield from self._serialize_cbr_rec_()

        yield 'END-OF-LOG:'

    def _serialize_cbr_rec_(self) -> typing.Iterator[str]:
        """Serialize the single QSOs for CBR"""
        for r in self.__qsos__:
            yield (f'QSO: {r.band.rjust(5)} {r.mode} {r.date} {r.time} '
                   f'{r.own_call.ljust(13)} {r.sent_rst.rjust(3)} {r.sent_exch.ljust(6)} '
                   f'{r.call.ljust(13)} {r.rcvd_rst.rjust(3)} {r.rcvd_exch.ljust(6)} {r.tx}')

    def process_points(self, rec: CBRRecord):
        # noinspection PyBroadException
        try:
            rated = 1
            points = rated
            if rec.call in self.__calls__:
                self.info('Duplicate QSO will not be rated')
                rated = 0
                points = 0
            else:
                self.__calls__.append(rec.call)
                self.__rated__ += 1
                self.__points__ = self.__rated__

            if BAND_FROM_CBR[rec.band] in self.__stats__:
                self.__stats__[BAND_FROM_CBR[rec.band]].qsos += 1
                self.__stats__[BAND_FROM_CBR[rec.band]].rated += rated
                self.__stats__[BAND_FROM_CBR[rec.band]].points += points
            else:
                self.__stats__[BAND_FROM_CBR[rec.band]] = BandStatistics(1, rated, points, 1, 0, 1)
        except Exception:
            self.exception()

    @classmethod
    def valid_bands(cls) -> tuple[CategoryBand, ...]:
        """Valid bands for the contest"""
        return CategoryBand.B_ALL,

    @classmethod
    def valid_modes(cls) -> tuple[CategoryMode, ...]:
        return CategoryMode.MIXED, CategoryMode.FM, CategoryMode.SSB, CategoryMode.CW

    @staticmethod
    def extract_exchange(exchange: str) -> ExchangeData | None:
        return ExchangeData(number=str(exchange))

    @staticmethod
    def prepare_exchange(exchange: ExchangeData) -> str:
        return exchange.number


class ContestLogEDI(ContestLog):
    """Base class for ham radio contests with an EDI result file (usually VHF/UHF and higher)"""
    contest_name = 'UKW-Contest'
    contest_exch_fmt = 'Number,Locator'

    def __init__(self, callsign: str, name: str, club: str, address: Address, email: str, locator: str,
                 band: type[CategoryBand], mode: type[CategoryMode],
                 pwr: type[CategoryPower] = CategoryPower.HIGH,
                 cat_operator: type[CategoryOperator] = CategoryOperator.SINGLE_OP,
                 assisted: type[CategoryAssisted] = CategoryAssisted.NON_ASSISTED,
                 tx: type[CategoryTransmitter] = CategoryTransmitter.ONE,
                 operators: list[str] = None, specific: str = '', skip_id: bool = False,
                 skip_warn: bool = False, logger=None, **params):
        super().__init__(callsign, name, club, address, email, locator,
                         band, mode, pwr, cat_operator,
                         assisted, tx, operators, specific, skip_id, skip_warn, logger)

        self.__qsos__: list[EDIRecord] = []

        self.__from_date__ = params.get('from_date', '2001-01-01')
        self.__to_date__ = params.get('to_date', '2001-01-01')
        self.__qth__ = params.get('qth', '<QTH>')
        self.__radio__ = params.get('radio', '<RADIO>')
        self.__power__ = params.get('pwr_watts', '<POWER_IN_WATTS>')
        self.__antenna__ = params.get('antenna', '<ANTENNA>')
        self.__ant_height_ground__ = params.get('ant_height_ground', '<ABOVE_GROUND>')
        self.__ant_height_sea__ = params.get('ant_height_sea', '<ABOVE_SEA>')

        self.__dok__ = specific

        self.__calls__ = []
        self.__locators__ = []
        self.__codxc__ = ['', '', 0]

        self.__edi_file__: typing.TextIO | None = None

    def check_band(self, adif_rec: dict[str, str]) -> bool:
        if adif_rec['BAND'].lower() != self.__header__['CATEGORY-BAND']:
            self.warning(
                f'Band "{adif_rec["BAND"]}" does not match with contest band {self.__header__["CATEGORY-BAND"]}')
            if self.__skip_warn__:
                return False
        return True

    def open_file(self, path: str = os.path.curdir):
        # EDI-Format from http://www.ok2kkw.com/ediformat.htm
        self.__edi_file__ = open(os.path.join(path, self.file_name), 'w')

        self.__edi_file__.write('[REG1TEST;1]\n')
        self.__edi_file__.write(f'TName={self.contest_name}\n')
        self.__edi_file__.write(f'TDate={self.__from_date__.replace("-", "")};{self.__to_date__.replace("-", "")}\n')

        # Operator
        self.__edi_file__.write(f'PCall={self.__header__["CALLSIGN"]}\n')
        self.__edi_file__.write(f'PWWLo={self.__header__["GRID-LOCATOR"].upper()}\n')
        self.__edi_file__.write(f'PExch={self.__header__["GRID-LOCATOR"].upper()}\n')
        self.__edi_file__.write(f'PAdr1={self.__qth__}\n')
        self.__edi_file__.write('PAdr2=\n')
        self.__edi_file__.write(f'PSect={self.__header__["CATEGORY-OPERATOR"]}\n')
        self.__edi_file__.write(f'PBand={BAND_MAP_EDI[self.__header__["CATEGORY-BAND"].lower()]}\n')
        self.__edi_file__.write(f'PClub={self.__header__["SPECIFIC"]}\n')

        # Responsible
        self.__edi_file__.write(f'RName={self.__header__["NAME"]}\n')
        self.__edi_file__.write(f'RCall={self.__header__["CALLSIGN"]}\n')

        addr = self.__header__['ADDRESS'].split('\n', 1)
        addr_2 = addr[1].replace('\n', ' ') if len(addr) > 1 else ''
        self.__edi_file__.write(f'RAdr1={addr[0]}\n')
        self.__edi_file__.write(f'RAdr2={addr_2}\n')

        self.__edi_file__.write(f'RPoCo={self.__header__["ADDRESS-POSTALCODE"]}\n')
        self.__edi_file__.write(f'RCity={self.__header__["ADDRESS-CITY"]}\n')
        self.__edi_file__.write(f'RCoun={self.__header__["ADDRESS-COUNTRY"]}\n')
        self.__edi_file__.write('RPhon=\n')
        self.__edi_file__.write(f'RHBBS={self.__header__["EMAIL"]}\n')

        # Operators
        self.__edi_file__.write(f'MOpe1={self.__header__["OPERATORS"]}\n')
        self.__edi_file__.write('MOpe2=\n')

        # Station
        self.__edi_file__.write(f'STXEq={self.__radio__}\n')
        self.__edi_file__.write(f'SPowe={self.__power__}\n')
        self.__edi_file__.write(f'SRXEq={self.__radio__}\n')
        self.__edi_file__.write(f'SAnte={self.__antenna__}\n')
        self.__edi_file__.write(f'SAntH={self.__ant_height_ground__};{self.__ant_height_sea__}\n')

    def summary(self) -> str:
        return (f'QSOs: {self.qsos}, Rated: {self.rated}, Points: {self.points}, '
                f'Claimed points: {self.claimed_points}')

    def build_record(self, adif_rec) -> EDIRecord | None:
        self.log.debug(adif_rec)

        if self.__header__['CATEGORY-BAND'] in ('all', adif_rec['BAND'].lower()):
            date = f'{adif_rec["QSO_DATE"][:4]}-{adif_rec["QSO_DATE"][4:6]}-{adif_rec["QSO_DATE"][6:8]}'
            if not self.__contest_date__:
                self.__contest_date__ = date

            srx_string = adif_rec['SRX_STRING'].upper().strip().replace(',', ' ').replace('_', ' ')
            if ' ' in srx_string:
                _, rloc = srx_string.split(' ', maxsplit=1)
            else:
                raise Exception(f'Problem with received exchange "{adif_rec["SRX_STRING"]}"')

            dist = 0
            try:
                loc = self.__header__['GRID-LOCATOR']
                dist = distance(loc if loc else adif_rec['MY_GRIDSQUARE'], rloc)
            except Exception as exc:
                self.exception(str(exc))

            if dist > self.__codxc__[2]:
                self.__codxc__ = [adif_rec['CALL'], rloc, dist]

            if rloc.upper() not in self.__locators__:
                self.__locators__.append(rloc.upper())

            return EDIRecord(BAND_MAP_CBR[adif_rec['BAND'].lower()],
                             MODE_MAP_CBR[adif_rec['MODE']],
                             date,
                             adif_rec['TIME_ON'][:4],
                             adif_rec['STATION_CALLSIGN'],
                             adif_rec['RST_SENT'],
                             f'{adif_rec["STX"]:03d}',
                             adif_rec['CALL'],
                             adif_rec['RST_RCVD'],
                             srx_string,
                             '0',
                             dist
                             )

        return None

    def write_records(self):
        if self.__edi_file__:
            # Contest info
            self.__edi_file__.write(f'CQSOs={len(self.__qsos__)};1\n')
            self.__edi_file__.write(f'CQSOP={self.claimed_points}\n')
            self.__edi_file__.write(f'CWWLs={len(self.__locators__)};0;1\n')
            self.__edi_file__.write('CWWLB=0\n')
            self.__edi_file__.write('CExcs=0;0;1\n')
            self.__edi_file__.write('CExcB=0\n')
            self.__edi_file__.write('CDXCs=0;0;1\n')
            self.__edi_file__.write('CDXCB=0\n')
            self.__edi_file__.write(f'CToSc={self.claimed_points}\n')
            self.__edi_file__.write(f'CODXC={self.__codxc__[0]};{self.__codxc__[1]};{self.__codxc__[2]}\n')

            # Remarks
            self.__edi_file__.write('[Remarks]\n')
            if 'SOAPBOX' in self.__header__ and self.__header__['SOAPBOX'].strip():
                self.__edi_file__.write(self.__header__['SOAPBOX'].strip() + '\n')

            # Records
            self.__edi_file__.write(f'[QSORecords;{len(self.__qsos__)}]\n')
            calls = []
            locators = []

            for rec in self.__qsos__:
                rnr, rloc = rec.rcvd_exch.split(' ')

                dup = False
                if rec.call not in calls:
                    calls.append(rec.call)
                else:
                    dup = True

                lnew = True
                if rloc.upper() not in locators:
                    locators.append(rloc.upper())
                else:
                    lnew = False

                edi_fields = [rec.date[2:].replace('-', ''),  # Date
                              rec.time[:4],  # Time
                              rec.call,  # Call
                              str(MODE_MAP_EDI[rec.mode]),  # Mode
                              rec.sent_rst,  # sent RST
                              rec.sent_exch,  # sent Exch
                              rec.rcvd_rst,  # rcvd RST
                              rnr,  # rcvd QSO nr
                              '',  # recvd Exch
                              rloc.upper(),  # Locator
                              str(rec.dist) if not dup else '0',  # QSO Points
                              '',  # New exch
                              'N' if lnew else '',  # New locator
                              '',  # New DXCC
                              '' if not dup else 'D'  # Duplicate QSO
                              ]
                self.__edi_file__.write(';'.join(edi_fields) + '\n')

    def close_file(self):
        if self.__edi_file__:
            self.__edi_file__.close()

    @property
    def claimed_points(self) -> int:
        return self.points

    def process_points(self, rec: EDIRecord):
        # noinspection PyBroadException
        try:
            rated = 1
            points = rec.dist
            if rec.call in self.__calls__:
                self.info('Duplicate QSO will not be rated')
                rated = 0
                points = 0
            else:
                self.__calls__.append(rec.call)
                self.__rated__ += 1
                self.__points__ += rec.dist

            if BAND_FROM_CBR[rec.band] in self.__stats__:
                self.__stats__[BAND_FROM_CBR[rec.band]].qsos += 1
                self.__stats__[BAND_FROM_CBR[rec.band]].rated += rated
                self.__stats__[BAND_FROM_CBR[rec.band]].points += points
            else:
                self.__stats__[BAND_FROM_CBR[rec.band]] = BandStatistics(1, rated, points, 1, 0, 1)
        except Exception:
            self.exception()

    @property
    def file_name(self) -> str:
        return (f'{self.__contest_date__}_{self.__header__["CALLSIGN"].replace("/", "-")}_'
                f'{self.contest_name.replace(" ", "-")}_{self.__header__["CATEGORY-BAND"]}.edi')

    @classmethod
    def valid_modes(cls) -> tuple[CategoryMode, ...]:
        return CategoryMode.MIXED, CategoryMode.FM, CategoryMode.SSB, CategoryMode.CW

    @classmethod
    def valid_bands(cls) -> tuple[CategoryBand, ...]:
        return (CategoryBand.B_6M, CategoryBand.B_4M, CategoryBand.B_2M, CategoryBand.B_70CM, CategoryBand.B_23CM,
                CategoryBand.B_9CM, CategoryBand.B_6CM, CategoryBand.B_3CM, CategoryBand.B_1_25CM, CategoryBand.B_6MM,
                CategoryBand.B_4MM, CategoryBand.B_2_5MM, CategoryBand.B_2MM, CategoryBand.B_1MM)

    @classmethod
    def valid_power(cls) -> tuple[CategoryPower, ...]:
        return CategoryPower.NONE,

    @staticmethod
    def extract_exchange(exchange: str) -> ExchangeData | None:
        exchange = exchange.upper().strip().replace(',', ' ').replace('_', ' ')
        if ' ' in exchange:
            r_nr, r_loc = exchange.split(' ', maxsplit=1)
            return ExchangeData(locator=r_loc.strip(), number=r_nr.strip())
        else:
            return None

    @staticmethod
    def prepare_exchange(exchange: ExchangeData) -> str:
        return f'{exchange.number},{exchange.locator}'



# Form implementation generated from reading ui file 'SelectWidget.ui'
#
# Created by: PyQt6 UI code generator 6.9.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_SelectWidget(object):
    def setupUi(self, SelectWidget):
        SelectWidget.setObjectName("SelectWidget")
        SelectWidget.resize(300, 150)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(SelectWidget.sizePolicy().hasHeightForWidth())
        SelectWidget.setSizePolicy(sizePolicy)
        self.horizontalLayout = QtWidgets.QHBoxLayout(SelectWidget)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.labelEnabled = QtWidgets.QLabel(parent=SelectWidget)
        self.labelEnabled.setObjectName("labelEnabled")
        self.verticalLayout_3.addWidget(self.labelEnabled)
        self.enabledListWidget = QtWidgets.QListWidget(parent=SelectWidget)
        self.enabledListWidget.setObjectName("enabledListWidget")
        self.verticalLayout_3.addWidget(self.enabledListWidget)
        self.horizontalLayout.addLayout(self.verticalLayout_3)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(parent=SelectWidget)
        self.label.setText("")
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.disablePushButton = QtWidgets.QPushButton(parent=SelectWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.disablePushButton.sizePolicy().hasHeightForWidth())
        self.disablePushButton.setSizePolicy(sizePolicy)
        self.disablePushButton.setMaximumSize(QtCore.QSize(20, 16777215))
        self.disablePushButton.setObjectName("disablePushButton")
        self.verticalLayout_2.addWidget(self.disablePushButton)
        self.enablePushButton = QtWidgets.QPushButton(parent=SelectWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.enablePushButton.sizePolicy().hasHeightForWidth())
        self.enablePushButton.setSizePolicy(sizePolicy)
        self.enablePushButton.setMaximumSize(QtCore.QSize(20, 16777215))
        self.enablePushButton.setObjectName("enablePushButton")
        self.verticalLayout_2.addWidget(self.enablePushButton)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_2.addItem(spacerItem1)
        self.horizontalLayout.addLayout(self.verticalLayout_2)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.labelDisabled = QtWidgets.QLabel(parent=SelectWidget)
        self.labelDisabled.setObjectName("labelDisabled")
        self.verticalLayout_4.addWidget(self.labelDisabled)
        self.disabledListWidget = QtWidgets.QListWidget(parent=SelectWidget)
        self.disabledListWidget.setObjectName("disabledListWidget")
        self.verticalLayout_4.addWidget(self.disabledListWidget)
        self.horizontalLayout.addLayout(self.verticalLayout_4)

        self.retranslateUi(SelectWidget)
        self.disablePushButton.clicked.connect(SelectWidget.disableItem) # type: ignore
        self.enablePushButton.clicked.connect(SelectWidget.enableItem) # type: ignore
        self.disabledListWidget.itemDoubleClicked['QListWidgetItem*'].connect(SelectWidget.enableItem) # type: ignore
        self.enabledListWidget.itemDoubleClicked['QListWidgetItem*'].connect(SelectWidget.disableItem) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(SelectWidget)

    def retranslateUi(self, SelectWidget):
        _translate = QtCore.QCoreApplication.translate
        SelectWidget.setWindowTitle(_translate("SelectWidget", "Form"))
        self.labelEnabled.setText(_translate("SelectWidget", "Items enabled"))
        self.disablePushButton.setText(_translate("SelectWidget", ">"))
        self.enablePushButton.setText(_translate("SelectWidget", "<"))
        self.labelDisabled.setText(_translate("SelectWidget", "Items disabled"))



import re
import logging

import requests
from adif_file import adi

from .Logger import Logger


class EQSLCommunicationException(Exception):
    pass


class EQSLRequestException(Exception):
    pass


class EQSLLoginException(Exception):
    pass


class EQSLUserCallMatchException(Exception):
    pass


class EQSLQSODuplicateException(Exception):
    pass


class EQSLADIFFieldException(Exception):
    pass


class EQSL:
    required_fields = ('QSO_DATE', 'TIME_ON', 'CALL', 'MODE', 'BAND')
    fields = required_fields + ('FREQ', 'QSLMSG', 'RST_SENT', 'MY_GRIDSQUARE', 'PROP_MODE', 'SUBMODE')
    image_pattern = re.compile(r'.*<img src="(.*)" alt="" />.*')
    upl_res_pattern = re.compile(r' *([EWCIR][a-z]*:.*)<BR>')

    def __init__(self, program: str, logger: Logger):
        self.__program_str__ = program

        self.log = logging.getLogger('EQSL')
        self.log.addHandler(logger)
        self.log.setLevel(logger.loglevel)
        self.logger = logger
        self.log.debug('Initialising...')

    def upload_log(self, username: str, password: str, record: dict) -> bool:
        if not username or not password:
            raise EQSLLoginException('Username or password missing')

        self._check_fields_(record)

        record = record.copy()

        for field in list(record.keys()):  # Create list object due to changes to dict below
            if not field in self.fields:
                record.pop(field)

        record['ADIF_VER'] = '3.1.4'
        record['PROGRAMID'] = self.__program_str__

        # Skip header and remove linebreaks
        data = adi.dumps({'RECORDS': [record]}).replace('\n', ' ')

        params = {
            'ADIFData': 'QSL-Upload ' + data,
            'EQSL_USER': username,
            'EQSL_PSWD': password,
        }

        r = requests.get('https://www.eQSL.cc/qslcard/importADIF.cfm', params=params)

        if r.status_code == 200:
            for res in re.findall(self.upl_res_pattern, r.text):
                if 'Error' in res:
                    if ('No match on eQSL_User/eQSL_Pswd' in res or
                            'Multiple accounts match eQSL_User/eQSL_Pswd' in res):
                        raise EQSLUserCallMatchException(res)
                    elif 'Missing eQSL_User' in res or 'Missing eQSL_Pswd' in res:
                        raise EQSLLoginException(res)
                    elif 'Missing ADIFData parameter' in res:
                        raise EQSLADIFFieldException(res)
                    else:
                        raise EQSLRequestException(res)
                elif 'Warning' in res:
                    if 'Bad record: Duplicate' in res:
                        raise EQSLQSODuplicateException(res)
                    else:
                        raise EQSLADIFFieldException(res)
                elif 'Caution' in res:
                    self.log.warning(f'eQSL result: {res}')
                elif 'Information' in res:
                    self.log.info(f'eQSL result: {res}')
                elif 'Result' in res:
                    self.log.debug(f'eQSL result: {res}')

            return True
        else:
            raise EQSLCommunicationException(f'eQSL error: HTTP-Error {r.status_code}')

    def _check_fields_(self, record: dict):
        for field in self.required_fields:
            if field not in record:
                raise EQSLADIFFieldException(field)

    def check_inbox(self, username: str, password: str, record: dict) -> str:
        if not username or not password:
            raise EQSLLoginException('Username or password missing')

        self._check_fields_(record)

        params = {
            'Username': username,
            'Password': password,
            'CallsignFrom': record['CALL'],
            'QSOYear': record['QSO_DATE'][:4],
            'QSOMonth': record['QSO_DATE'][4:6],
            'QSODay': record['QSO_DATE'][6:],
            'QSOHour': record['TIME_ON'][:2],
            'QSOMinute': record['TIME_ON'][2:4],
            'QSOBand': record['BAND'],
            'QSOMode': record['MODE'],
        }

        r = requests.get('https://www.eQSL.cc/qslcard/GeteQSL.cfm', params=params)
        if r.status_code == 200:
            if r.text.strip().startswith('Error'):
                if ('No match on Username/Password for that QSO Date/Time' in r.text or
                        'overlapping accounts for that QSO Date/Time' in r.text):
                    raise EQSLUserCallMatchException(r.text.strip())
                elif ('User is Regular member but must be at least Silver to download mass eQSLs.' in r.text or
                      'User is Bronze member but must be at least Silver to download mass eQSLs.' in r.text or
                      'Not Authorized to download mass eQSLs.' in r.text):
                    raise EQSLLoginException(r.text.strip())
                else:
                    raise EQSLRequestException(r.text.strip())
            elif 'Variable PASSWORD is undefined.' in r.text:
                raise EQSLLoginException('Variable PASSWORD is undefined')
            else:
                url_part = re.findall(self.image_pattern, r.text)
                if url_part:
                    return 'https://www.eQSL.cc' + url_part[0]
                else:
                    return ''
        else:
            self.log.debug(f'Status code: {r.status_code}')
            raise EQSLCommunicationException(f'eQSL error: HTTP-Error {r.status_code}')

    @staticmethod
    def receive_qsl_card(url: str) -> bytes:
        if url:
            r = requests.get(url)

            if r.status_code == 200:
                return r.content
            else:
                raise EQSLCommunicationException(f'eQSL error: HTTP-Error {r.status_code}')
        else:
            return b''
